/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "ucs.h"
#include "c-lib.h"
#include "smemory.h"
#include "str-lib.h"

#include "dafstype.h"
#include "entity.h"
#include "dafs.h"
#include "utf.h"

#ifndef min
#define min(a,b) ( (a)<(b) ? (a):(b))
#define max(a,b) ( (a)>(b) ? (a):(b))
#endif

static int 
isParent(iEntityPtr par, iEntityPtr ch)
{
	iEntityPtr cur;
	
	for(cur=i_GetParentEntity(ch);cur;cur=i_GetParentEntity(cur))
		if (par == cur)
			return 1;
	return 0;
}

static void 
deleteChildren(iEntityPtr en)
{
	iEntityPtr cur;
	
	for(cur=i_GetChildEntity(en);cur;cur=i_GetChildEntity(en)){
		i_DisposeEntity(cur);
	}
}

iDAFSError 
i_Collapse(iEntityPtr en,iUCode *nonrec)
{
	iEntityPtr cur;
	int num,i;
	const char *s;
	iDAFSError e=DAFSOK;
	iCallBackProc save=0;
	iUCode nr,c;
	
	if(gCallBack){
		MatrCallBack(mWillUpdate,gUserData,en);
		save=gCallBack;
		gCallBack=0;
	}
	nr=(nonrec) ? *nonrec : '^';
	cur=i_FindEntityForward(en,0,DATAEN,0);
	while(cur && isParent(en,cur)){
		switch(i_GetEDataType(cur)){
		case iDPosSet:
		case iD1PosSet:
			e=i_AddUni(en,-1,(iUCode) ((c=i_GetGlyph(en,0)) ? c : nr));
			if (e)
				return e;
			break;
		case iDLong:
			num=i_GetInt(cur,0);
			for(i=0;i<num;i++) {
				if(i_GetEntityType(cur)==iSpaceT)
					e=i_AddUni(en,-1,' ');
				else if(i_GetEntityType(cur)==iNewLineT)
					e=i_AddUni(en,-1,'\n');
				if (e)
					return e;
			}
			break;
		case iDString:
			for(i=0;i<strlen(s=i_GetText(cur));i++)
				e=i_AddUni(en,-1,(short)s[i]);
			if (e)
				return e;
			break;
		case iDUniString:
			e=i_AddUnis(en,-1,(iUCode*)i_GetUniText(cur));
			if (e)
				return e;
			break;
		case iDNo:
			break;
		default :
			return UnknownDAFSError;
			break;
		}
		cur=i_FindEntityForward(cur,0,DATAEN,0);
	}	
	if (save) {
		gCallBack=save;
		MatrCallBack(mUpdate,gUserData,en);
	}
	deleteChildren(en);
	return DAFSOK;
}

static iDAFSError 
MergeEntity(iEntityPtr l,iEntityPtr r,
			iEntityPtr *bp,int *beg,iEntityPtr *ep,int *end)
{
	int t,len;
	iDAFSError e;
	iEntityPtr next,cur,temp;
	const char *text;
	const iUCode *utext;
	iBox bl,br, box;
	
	t = i_GetEDataType(l);
	switch (t) {
	case iDString:
		text = i_GetText(l);
		if (r == *ep)
			*end += strlen(text);
		*beg = strlen(text);
		*bp = r;
		if (e = i_AddChars(r,0,(char*)text))
			return e;
		bl = i_GetBounds(l);
		br = i_GetBounds(r);
		if (bl.h && br.h){
			i_UnionBox(bl, br, &box);
			i_SetBounds(r,box);
		}
		i_DisposeEntity(l);
		break;
	case iDUniString:
		utext = i_GetUniText(l);
		if (r == *ep)
			*end += ucslen(utext);
		*beg = ucslen(utext);
		*bp = r;
		if (e = i_AddUnis(r,0,(iUCode*)utext))
			return e;
		bl = i_GetBounds(l);
		br = i_GetBounds(r);
		if (bl.h && br.h){
			i_UnionBox(bl, br, &box);
			i_SetBounds(r,box);
		}
		i_DisposeEntity(l);
		break;
	case iDLong:
		len = i_GetInt(l,0);
		if (r == *ep)
			*end += len;
		*beg = len;
		*bp = r;
		i_SetInt(r,len + i_GetInt(r,0), 0);
		bl = i_GetBounds(l);
		br = i_GetBounds(r);
		if (bl.h && br.h){
			i_UnionBox(bl, br, &box);
			i_SetBounds(r,box);
		}
		i_DisposeEntity(l);
		break;
	case iDNo:
		if (!i_GetChildEntity(r) || !i_GetChildEntity(l))
			break;
		for (cur = i_GetChildEntity(r);cur;cur = next) {
			next = i_GetNextEntity(cur);
			temp = i_GetLastEntity(l);
			bl = i_GetBounds(l);
			br = i_GetBounds(cur);
			if (bl.h && br.h){
				i_UnionBox(bl, br, &box);
				i_SetBounds(l,box);
			}
			if ((e = i_TransferEntity(cur,l)))
				return e;
			if ((i_GetEntityType(temp) == i_GetEntityType(cur)) &&
				(i_GetEDataType(temp) == i_GetEDataType(cur)) &&
				(e = MergeEntity(temp,cur,bp,beg,ep,end)))
				return e;
		}
		bl = i_GetBounds(l);
		br = i_GetBounds(r);
		if (bl.h && br.h){
				i_UnionBox(bl, br, &box);
				i_SetBounds(l,box);
		}
		i_DisposeEntity(r);
		break;
	case iDPosSet:
	case iD1PosSet:
		break;
	}
	return DAFSOK;
}

static iDAFSError 
killachar(iEntityPtr *bp,int *beg,
		  iEntityPtr *ep,int *end,
		  int merge)
{
	int et,len;
	iEntityPtr par,np,prev,next;
	iDAFSError e;
	
	et = i_GetEDataType(*bp);
	if (et == iDString) {
		len = strlen(i_GetText(*bp));
		if (len > 1) {
			len--;
			i_DelChar(*bp,*beg);
			if (*bp == *ep)
				(*end)--;
			else if (len == *beg) {
				*beg = 0;
				*bp = i_FindEntityForward(*bp,0,DATAEN,0);
			}
			return DAFSOK;
		}
	} else if (et == iDUniString) {
		len = ucslen(i_GetUniText(*bp));
		if (len > 1) {
			len--;
			i_DelUni(*bp,*beg);
			if (*bp == *ep)
				(*end)--;
			else if (len == *beg) {
				*beg = 0;
				*bp = i_FindEntityForward(*bp,0,DATAEN,0);
			}
			return DAFSOK;
		}
	} else if (et == iDLong) {
		len = i_GetInt(*bp,0);
		if (len > 1) {
			i_SetInt(*bp,--len,0);
			if (*bp == *ep)
				(*end)--;
			else if (len == *beg) {
				*beg = 0;
				*bp = i_FindEntityForward(*bp,0,DATAEN,0);
			}
			return DAFSOK;
		}
	}

    /* it is OK to set *bp to an */
    /* entity with no data because kill a char will be */
    /* called until *bp == *ep */
	np = i_FindEntityForward(*bp,0,DATAEN,0); 
	for (;;) {				
		next = i_GetNextEntity(*bp);	
		prev = i_GetPrevEntity(*bp);
		if (next || prev) break;
		par = i_GetParentEntity(*bp);
		i_DisposeEntity(*bp);
		*bp = par;
	}
	i_DisposeEntity(*bp);
	*bp = np;
	*beg = 0;
	if (merge && prev && next &&
			i_GetEDataType(next) == i_GetEDataType(prev) &&
#if 1
		    i_GetEDataType(next) != iDNo &&
#endif
			(e = MergeEntity(prev,next,bp,beg,ep,end)))
		return e;
	return DAFSOK;
}

static iBox 
fixBounds(iEntityPtr en)
{
	iEntityPtr cur;
	iBox nb,b,tb;

	b = i_GetBounds(en);
	if (!(cur=i_GetChildEntity(en))) return b;
	nb = fixBounds(cur);
	cur = i_GetNextEntity(cur);
	for (;cur;cur = i_GetNextEntity(cur)) {
		tb = fixBounds(cur);
		i_UnionBox(tb,nb,&nb);
	}
	return nb;
}

static int 
getPrntBox(iEntityPtr en,iBox *bp)
{
	en = i_GetParentEntity(en);
	if (!en) return 0;
	*bp = i_GetBounds(en);
	return 1;
}

static void 
guessBox(iEntityPtr new_entity)
{
	iEntityPtr prev,next;
	iBox b,nb,pb,prb;
	int nlp,nln,t,nt;
	
	nt = i_GetEntityType(new_entity);
	nlp = 0;
	for (prev = i_FindEntityBack(new_entity,0,DATAEN,0);prev;prev = i_FindEntityBack(prev,0,DATAEN,0)) {
		t = i_GetEntityType(prev);
		if (t == iNewLineT) nlp += i_GetInt(prev,0);
		else break;
	}
	nln = 0;
	for (next = i_FindEntityForward(new_entity,0,DATAEN,0);next;next = i_FindEntityBack(next,0,DATAEN,0)) {
		t = i_GetEntityType(next);
		if (t == iNewLineT) nln += i_GetInt(next,0);
		else break;
	}
	if (next) nb = i_GetBounds(next);
	else nb.w = nb.h = 0;
	if (prev) pb = i_GetBounds(prev);
	else pb.w = pb.h = 0;
	if (nb.w && nb.h && pb.w && pb.h) {
		if (!nln && !nlp) {
			b.y = min(pb.y,nb.y);
			b.h = max(pb.y+pb.h,nb.y+nb.h) - b.y;
			b.x = pb.x + pb.w;
			b.w = nb.x - b.x;
		} else if (!nln) {
			b = nb;
			if (getPrntBox(new_entity,&prb))
				b.x = prb.x;
			else
				b.x = nb.x - (b.h * 2) / 3;
			b.w = nb.x - b.x;
		} else if (!nlp) {
			b = pb;
			b.x = pb.x + pb.w;
			if (getPrntBox(new_entity,&prb))
				b.w = (prb.x + prb.w) - b.x;
			else
				b.w = (b.h * 2) / 3;
		} else {
			if (!getPrntBox(new_entity,&b))
				b = pb;
			b.h = (nb.y - pb.y) / (nln + nlp);
			b.y += b.h * nlp;
		}
	} else if (pb.w && pb.h) {
		if (nlp) {
			if (!getPrntBox(new_entity,&b))
				b = pb;
			else b.h = pb.h;
			b.y += (b.h + (b.h / 8)) * nlp;
		} else {
			b = pb;
			b.x = pb.x + pb.w;
			if (getPrntBox(new_entity,&prb))
				b.w = (prb.x + prb.w) - b.x;
			else
				b.w = (b.h * 2) / 3;
		}
	} else if (nb.w && nb.h) {
		if (nln) {
			if (!getPrntBox(new_entity,&b))
				b = nb;
			else b.h = nb.h;
			b.y -= (b.h + (b.h / 8)) * nln;
		} else {
			b = nb;
			if (getPrntBox(new_entity,&prb))
				b.x = prb.x;
			else
				b.x = nb.x - (b.h * 2) / 3;
			b.w = nb.x - b.x;
		}
	}
	if (b.w <= 0) b.w = 1;
	if (b.h <= 0) b.h = 1;
	i_SetBounds(new_entity,b);
}

static iDAFSError 
splitEntity(iEntityPtr bp,iEntityPtr p,iTypeIndex type,iEntityPtr *newp)
{
	iEntityPtr next,par,new,lastnew;
	iDAFSError e;
	
	while (!i_GetPrevEntity(bp) && bp != p)
		bp = i_GetParentEntity(bp);
	if (i_GetParentEntity(bp) == i_GetParentEntity(p)) {
		new = i_NewEntityBefore(type,bp);
		if (!new) return MallocFailError;
		if (newp) *newp = new;
		return DAFSOK;
	}
	lastnew = 0;
	for (par = bp;par != p;) {
		if (bp) par = i_GetParentEntity(bp);
		else par = i_GetParentEntity(par);
		new = i_NewEntityAfter(i_GetEntityType(par),p);
		if (!new) return MallocFailError;
		if (lastnew && (e = i_TransferEntity(lastnew,new)))
			return e;
		for (;bp;bp = next) {
			next = i_GetNextEntity(bp);
			if (e = i_TransferEntity(bp,new))
				return e;
		}
		i_SetBounds(new,fixBounds(new));
		i_SetBounds(par,fixBounds(par));
		i_DisposeProp(par,"Flag");
		bp = i_GetNextEntity(par);
		lastnew = new;
	}
	new = i_NewEntityAfter(type,p);
	if (!new) return MallocFailError;
	i_MoveAfter(lastnew,new);
	guessBox(new);
	if (newp) *newp = new;
	return DAFSOK;
}

static iDAFSError 
deleteSelection(iEntityPtr *bp, int *beg, iEntityPtr *ep, int *end,
	   int merge) 
{
	iDAFSError e;
	
	while (*bp != *ep || *beg != *end)
		if (e = killachar(bp,beg,ep,end,merge))
			return e;
	return DAFSOK;
}

iDAFSError 
i_DeleteSelection(iEntityPtr *bp, int *beg, iEntityPtr *ep, int *end) 
{
	return deleteSelection(bp,beg,ep,end,1);
}

static void 
splitBoxes(iEntityPtr new_entity,iEntityPtr bp,int spot,int len)
{
	iBox b,bb;
	
	b = bb = i_GetBounds(bp);
	b.w = (bb.w * spot) / len;
	bb.x += b.w;
	bb.w -= b.w;
	i_SetBounds(new_entity,b);
	i_SetBounds(bp,bb);
}

static void
myAddChar(iEntityPtr p, int len, iUCode key)
{
	/* ah ha --can't add to the ascii string a non-ascii value */
			/* but aren't there more guys like this? */
	if(key > 127){
		iUCode *us;
		us = i_AllocUniText(p);
		i_SetUniText(p, us);
		smfree(us);
		i_AddUni(p, len, key);
	} else {
		i_AddChar(p, len, key);
	}
}

iDAFSError 
i_InsertChar(iEntityPtr *bp, int *beg, 
			 iEntityPtr *ep, int *end, 
			 iUCode key) 
{
	iDAFSError e;
	iEntityPtr p,new_entity;
	int pdt,dt,len;
	char *temp;
	iUCode *tempu;
	short usave; 
	char csave;
	iTypeIndex t,pt;
	p = i_FindEntityBack(*bp,0,DATAEN,0);
	if (p) {
		pt = i_GetEntityType(p);
		pdt = i_GetEDataType(p);
	} else {
		pt = iNullT;
		pdt = iDNo;
	}
	t = i_GetEntityType(*bp);
	dt = i_GetEDataType(*bp);
	switch (key) {
	case '\b':
		break;
	case '\n':
	case '\r':
		if (t == iNewLineT) {
			i_SetInt(*bp,i_GetInt(*bp,0) + 1,0);
			(*beg)++;
			if (*bp == *ep) (*end)++;
		} else if (*beg) {
			if (t == iSpaceT) {
				new_entity = i_NewEntityBefore(t,*bp);
				if (!new_entity) return MallocFailError;
				len = i_GetInt(*bp,0);
				i_SetInt(new_entity,*beg,0);
				i_SetInt(*bp,len - *beg,0);
				splitBoxes(new_entity,*bp,*beg,len);
				*beg = 0;
				if (*bp == *ep) *end = 0;
				if (e = splitEntity(*bp,new_entity,iNewLineT,&new_entity))
					return e;
				i_SetInt(new_entity,1,0);
			} else if (dt == iDString) {
				new_entity = i_NewEntityBefore(t,*bp);
				if (!new_entity) return MallocFailError;
				len = strlen(i_GetText(*bp));
				temp = smalloc(len + 1);
				strcpy(temp,i_GetText(*bp));
				csave = temp[*beg];
				temp[*beg] = 0;
				i_SetText(new_entity,temp);
				temp[*beg] = csave;
				i_SetText(*bp,&temp[*beg]);
				smfree(temp);
				splitBoxes(new_entity,*bp,*beg,len);
				*beg = 0;
				if (*bp == *ep) *end = 0;
				p = i_FindEntityBack(*bp,iLineT,ALL,0);
				if (!p) {
					new_entity = i_NewEntityBefore(iNewLineT,*bp);
					if (!new_entity) return MallocFailError;
				} else
					if (e = splitEntity(*bp,p,iNewLineT,&new_entity))
						return e;
				i_SetInt(new_entity,1,0);
			} else if (dt == iDUniString) {
				new_entity = i_NewEntityBefore(t,*bp);
				if (!new_entity) return MallocFailError;
				len = ucslen(i_GetUniText(*bp));
				tempu = (iUCode*)smalloc((len + 1) * sizeof(iUCode));
				ucscpy(tempu,i_GetUniText(*bp));
				usave = tempu[*beg];
				tempu[*beg] = 0;
				i_SetUniText(new_entity,tempu);
				tempu[*beg] = usave;
				i_SetUniText(*bp,&tempu[*beg]);
				smfree(tempu);
				splitBoxes(new_entity,*bp,*beg,len);
				*beg = 0;
				if (*bp == *ep) *end = 0;
				p = i_FindEntityBack(*bp,iLineT,ALL,0);
				if (!p) {
					new_entity = i_NewEntityBefore(iNewLineT,*bp);
					if (!new_entity) return MallocFailError;
				} else
					if (e = splitEntity(*bp,p,iNewLineT,&new_entity))
						return e;
				i_SetInt(new_entity,1,0);
			} else {
				p = i_FindEntityBack(*bp,iLineT,ALL,0);
				if (!p) {
					new_entity = i_NewEntityBefore(iNewLineT,*bp);
					if (!new_entity) return MallocFailError;
				} else
					if (e = splitEntity(*bp,p,iNewLineT,&new_entity))
						return e;
				i_SetInt(new_entity,1,0);
			}
		} else if (pt == iNewLineT) {
			i_SetInt(p,i_GetInt(p,0) + 1,0);
		} else {
			if (i_GetEntityType(*bp) == iLineT) p = 0;
			else p = i_FindEntityBack(*bp,iLineT,ALL,0);
			if (!p) {
				new_entity = i_NewEntityBefore(iNewLineT,*bp);
				if (!new_entity) return MallocFailError;
			} else
				if (e = splitEntity(*bp,p,iNewLineT,&new_entity))
					return e;
			i_SetInt(new_entity,1,0);
		}
		break;
	case ' ':
		if (t == iSpaceT) {
			i_SetInt(*bp,i_GetInt(*bp,0) + 1,0);
			(*beg)++;
			if (*bp == *ep) (*end)++;
		} else if (*beg) {
			if (t == iNewLineT) {
				new_entity = i_NewEntityBefore(iNewLineT,*bp);
				if (!new_entity) return MallocFailError;
				i_SetInt(new_entity,*beg,0);
				i_SetInt(*bp,i_GetInt(*bp,0) - *beg,0);
				*beg = 0;
				if (*bp == *ep) *end = 0;
				new_entity = i_NewEntityBefore(iLineT,*bp);
				if (!new_entity) return MallocFailError;
				i_AddUni(new_entity,0,key);
			} else if (dt == iDString) {
				myAddChar(*bp,*beg,key);
				(*beg)++;
				if (*bp == *ep) (*end)++;
			} else if (dt == iDUniString) {
				i_AddUni(*bp,*beg,key);
				(*beg)++;
				if (*bp == *ep) (*end)++;
			}
		} else if (pt == iSpaceT) {
			i_SetInt(p,i_GetInt(p,0) + 1,0);
		} else if (dt == iDString) {
			myAddChar(*bp,0,key);
			(*beg)++;
			if (*bp == *ep) (*end)++;
		} else if (dt == iDUniString) {
			i_AddUni(*bp,0,key);
			(*beg)++;
			if (*bp == *ep) (*end)++;
		} else if (pdt == iDString) {
			myAddChar(p,strlen(i_GetText(p)),key);
		} else if (pdt == iDUniString) {
			i_AddUni(p,ucslen(i_GetUniText(p)),key);
		} else if (t == iNewLineT) {
			new_entity = i_NewEntityBefore(iLineT,*bp);
			if (!new_entity) return MallocFailError;
			i_AddUni(new_entity,0,key);
		} else {
			p = i_FindEntityBack(*bp,iLineT,ALL,0);
			if (e = splitEntity(*bp,p,iLineT,&new_entity))
				return e;
			i_AddUni(new_entity,0,key);
		}
		break;
	default:
		if (*beg) {
			if (t == iSpaceT) {
				new_entity = i_NewEntityBefore(iSpaceT,*bp);
				if (!new_entity) return MallocFailError;
				len = i_GetInt(*bp,0);
				i_SetInt(new_entity,*beg,0);
				i_SetInt(*bp,len - *beg,0);
				splitBoxes(new_entity,*bp,*beg,len);
				*beg = 0;
				if (*bp == *ep) *end = 0;
				p = i_FindEntityBack(*bp,iLineT,ALL,0);
				if (e = splitEntity(*bp,p,iLineT,&new_entity))
					return e;
				i_AddUni(new_entity,0,key);
			} else if (t == iNewLineT) {
				new_entity = i_NewEntityBefore(iNewLineT,*bp);
				if (!new_entity) return MallocFailError;
				i_SetInt(new_entity,*beg,0);
				i_SetInt(*bp,i_GetInt(*bp,0) - *beg,0);
				*beg = 0;
				if (*bp == *ep) *end = 0;
				new_entity = i_NewEntityBefore(iLineT,*bp);
				if (!new_entity) return MallocFailError;
				i_AddUni(new_entity,0,key);
				guessBox(new_entity);
			} else if (dt == iDString) {
				myAddChar(*bp,*beg,key);
				(*beg)++;
				if (*bp == *ep) (*end)++;
			} else if (dt == iDUniString) {
				i_AddUni(*bp,*beg,key);
				(*beg)++;
				if (*bp == *ep) (*end)++;
			}
		} else if (dt == iDString) {
			myAddChar(*bp,0,key);
			(*beg)++;
			if (*bp == *ep) (*end)++;
		} else if (dt == iDUniString) {
			i_AddUni(*bp,0,key);
			(*beg)++;
			if (*bp == *ep) (*end)++;
		} else if (pdt == iDString) {
			myAddChar(p,strlen(i_GetText(p)),key);
		} else if (pdt == iDUniString) {
			i_AddUni(p,ucslen(i_GetUniText(p)),key);
		} else if (t == iNewLineT) {
			new_entity = i_NewEntityBefore(iLineT,*bp);
			if (!new_entity) return MallocFailError;
			i_AddUni(new_entity,0,key);
			guessBox(new_entity);
		} else {
			p = i_FindEntityBack(*bp,iLineT,ALL,0);
			if (e = splitEntity(*bp,p,iLineT,&new_entity))
				return e;
			i_AddUni(new_entity,0,key);
			guessBox(new_entity);
		}
		break;
	}
	return DAFSOK;
}

iDAFSError 
i_Edit(iEntityPtr *bp, int *beg, 
	   iEntityPtr *ep, int *end, 
	   iUCode key) 
{
	iDAFSError e;
	iEntityPtr p;
	
	if (key == '\b' && *bp == *ep && *beg == *end) {
		if (*beg) (*beg)--;
		else {
			p = i_FindEntityBack(*bp,0,DATAEN,0);
			if (!p) return DAFSOK;
			*bp = p;
			switch (i_GetEDataType(*bp)) {
			case iDString:
				*beg = strlen(i_GetText(*bp))-1;
				break;
			case iDUniString:
				*beg = ucslen(i_GetUniText(*bp))-1;
				break;
			case iDLong:
				*beg = i_GetInt(*bp,0)-1;
				break;
			case iDPosSet:
			case iD1PosSet:
				*beg = 0;
				break;
			default: break;
			}
		}
	}
	if (e = deleteSelection(bp,beg,ep,end,key == '\b'))
		return e;
	return i_InsertChar(bp,beg,ep,end,key);
}

static int isDataEnSub(iEntityPtr target)
{
	return (
		(i_GetEDataType(target) != iDNo) && 
		(i_GetEDataType(target) != iDConf) && 
		(i_GetEDataType(target) != iDUserData) &&
		((i_GetEDataType(target) != iDLong) || target->data->type == iSpaceT || target->data->type == iNewLineT)
		);
}


int 
_iAncestorIsData(iEntityPtr en)
{
	iEntityPtr parent = en ? i_GetParentEntity(en) : 0;
	int value;
	
	value =  (parent ? 
			  isDataEnSub(parent) || _iAncestorIsData(parent) : 
			  FALSE);
	
	return value;
}

int _iIsDataEn(iEntityPtr target)
{
	return isDataEnSub(target) && !_iAncestorIsData(target);
}

void i_Object2Range(iEntityPtr e,iEntityPtr *bp,int *beg,iEntityPtr *ep,int *end)
{
	iEntityPtr t;
	
	*beg=0;
	*end=0;	
	if (_iIsDataEn(e)) {
		*bp=e;
		*ep=i_FindEntityForward(e,0,DATAEN,0);
	}else{
		*bp=i_FindEntityForward(e,0,DATAEN,0);
		*ep=e;
		while(t=i_GetLastEntity(*ep))
			*ep=t;
		*ep=i_FindEntityForward(*ep,0,DATAEN,0);
	} 
	return;
}

iEntityPtr  
i_Range2Object(iEntityPtr bp,int beg,iEntityPtr ep,int end)
{
	iEntityPtr a,b;
	iBox bx;
	
	if (bp==ep) {
		bx=i_GetBounds(bp);
		if (bx.w)
			return bp;
		if(beg==end&&end==0){
			bp=i_FindEntityBack(ep,0,DATAEN,0);
			if(!bp)
				bp=ep;
			else
				ep=bp;
		}
	}
	if(end==0)
		ep=i_FindEntityBack(ep,0,DATAEN,0);
	
	for(a=bp;a;a=i_GetParentEntity(a))
		for(b=ep;b;b=i_GetParentEntity(b))
			if(a==b){
				bx=i_GetBounds(a);
				if(bx.w!=0)
					return a;
			}
	return 0;
}

static iDAFSError 
fixalloc(iUCode **text,int *pos,int nl,int *len)
{
	iUCode *nt;
	if ((nl + *pos) >= *len) {
		(*len) += nl + 100;
		nt = (iUCode*)smrealloc(*text,sizeof(iUCode) * (*len));
		if (!nt) return MallocFailError;
		*text = nt;
	}
	return DAFSOK;
}

static iDAFSError 
GetAllUniText(iEntityPtr en,iUCode **text,int *pos,int *len,iUCode *nonrec)
{
	iEntityPtr cur;
	int nc;
	char *s;
	iDAFSError e;
	iUCode *pt;
	iUCode c,nr;
	int t;
	nr=(nonrec) ? *nonrec : '^';
	if(!i_GetStringProp(en,"Alias",&s)){
		t=strlen(s);
		if (e = fixalloc(text,pos,t,len))
			return e;
		pt = (*text) + *pos;
		while (*s) *pt++ = *s++;
		(*pos) += t;
		*pt = 0;
		return DAFSOK;
	}
	t = i_GetEntityType(en);
	if (t == iSpaceT || t == iNewLineT) {
		nc = (t == iSpaceT? ' ': '\n');
		t = i_GetInt(en,0);
		if (e = fixalloc(text,pos,t,len))
			return e;
		pt = (*text) + *pos;
		(*pos)+=t;
		while (t--) *pt++ = nc;
		*pt = 0;
		return DAFSOK;
	}
	cur = i_GetChildEntity(en);
	if (!cur) {
		t = i_GetEDataType(en);
		if (t == iDNo) return DAFSOK;
		if (t == iDString) {
			s=(char*)i_GetText(en);
			t=strlen(s);
			if (e = fixalloc(text,pos,t,len))
				return e;
			pt = (*text) + *pos;
			while (*s) *pt++ = *s++;
			(*pos) += t;
			*pt = 0;
			return DAFSOK;
		}
		if (t == iDUniString) {
			pt=(iUCode*)i_GetUniText(en);
			t=ucslen(pt);
			if (e = fixalloc(text,pos,t,len))
				return e;
			ucscat((*text) + *pos,pt);
			(*pos) += t;
			return DAFSOK;
		}
		if (e = fixalloc(text,pos,1,len))
			return e;
		(*text)[(*pos)++] = (c=i_GetGlyph(en,0)) ? c : nr;
		(*text)[*pos] = 0;
		return DAFSOK;
	}
	t = i_GetChildType(en) == iAnd;
	for (; cur;cur = i_GetNextEntity(cur)) {
		if (e = GetAllUniText(cur,text,pos,len,nonrec))
			return e;
		if (!t) break;
	}
	return DAFSOK;
}

iUCode *
i_AllocUniText(iEntityPtr en)
{
	int len = 100,pos = 0;
	iUCode *text = (iUCode*)smalloc(len * sizeof(iUCode));
	iDAFSError e;
	
	if (!text) return 0;
	*text = 0;
	e = GetAllUniText(en,&text,&pos,&len,0);
	if (e) {
		smfree(text);
		text = 0;
	}
	return text;
}

void 
i_FreeUniText(iUCode *text)
{
	smfree(text);
}

#define UA_NUM  25
#define UA_SIZE 50

/*
// bogus this is odd -- typedefs are defined elsewhere,
// but the structures are defined here
// why bother
*/

struct iUAEntry {
	int index;
	char string[UA_SIZE];
};

struct iUATable {
	int num;
	iUAEntry entries[UA_NUM];
};


static int 
tableCmp(iUAEntryPtr a, iUAEntryPtr b)
{	
	return a->index - b->index;
}

iDAFSError 
i_ReadUATable(iUATablePtr *table, char *filename)
{
	FILE *fp;
	int i, alloc_size = UA_NUM;
	char line[256],*p;

	/* allocate space for the table */
	*table = (iUATablePtr) smalloc(sizeof(iUATable));
	if(!*table) return MallocFailError;

	/* read in filter */
	if(!(fp = fopen(filename, "rt")))
		return ReadError;

	/* process the filter table */
	i = 0;
	while(!feof(fp))
		if(fgets(line, sizeof line, fp) && strlen(line) > 2){
			switch(line[1]){
			  case 'x':
				sscanf(line, "0x%x%*[ \t\"]%s", &(*table)->entries[i].index, (*table)->entries[i].string);
				break;
			  default:
				sscanf(line, "%d%*[ \t\"]%s", &(*table)->entries[i].index,(char*) (*table)->entries[i].string);
			}
			if(p = (char*)strrchr((char*)(*table)->entries[i].string,'"')) *p = (char)'\0';
			i++;
			if(i == alloc_size){
				if(!(*table = smrealloc(*table, sizeof(iUATable) + sizeof(iUAEntry) * (alloc_size))))
					return MallocFailError;
				alloc_size += UA_NUM;
			}
		}
	(*table)->num = i;

	/* use quick sort to put table in order */
	qsort((*table)->entries, (*table)->num, sizeof (iUAEntry), (int(*)(const void *, const void *))tableCmp);

	return DAFSOK;
}

void 
i_DisposeUATable(iUATablePtr table)
{
	smfree(table);
} 

char *
i_AllocASCIIText(iEntityPtr en, iUATablePtr table)
{
	int len = 100,pos = 0,i;
	iUCode *text = (iUCode*)smalloc(len * sizeof(iUCode));
	char *ascii_text,*p;
	iDAFSError e;
	
	if(!en)
		return 0;
	if (!text) return 0;
	*text = 0;
	e = GetAllUniText(en,&text,&pos,&len,0);
	if (e) {
		smfree(text);
		return 0;
	}
	ascii_text=smalloc((pos + 1) *sizeof(char));
	if (!ascii_text) {
		smfree(text);
		return 0;
	}
	len = pos;
	/* convert uni to ascii */
	if(table){
		int i, j = 0, size = pos;
		iUAEntry key;
		iUAEntryPtr entry; 

		for(i = 0; i < pos; i++){
			key.index = text[i];
			if(entry = (iUAEntryPtr) bsearch(&key, table->entries, table->num, sizeof(iUAEntry), 
											 (int(*)(const void *, const void *)) tableCmp)){
				size += strlen(entry->string);
				if ((j+size) >= len) {
					len += size * 4;
					if(!(p = smrealloc(ascii_text, len+1))) {
						smfree(text);
						smfree(ascii_text);
						return 0;
					}
					ascii_text = p;
				}
				strcpy(ascii_text+j, entry->string);
				j = strlen(ascii_text);
			} else ascii_text[j++] = (char) text[i];
		}
		ascii_text[j] = 0;
	}
	else {
		for (i=0; i<pos; i++)
			ascii_text[i] = (char) text[i];
		ascii_text[i] = 0;
	}

	smfree(text);
	return ascii_text;
}

void 
i_FreeASCIIText(char *text)
{
	smfree(text);
}

iDAFSError 
i_StartObjEdit(iEntityPtr ent,iEditRecordPtr *r)
{
	*r=(iEditRecordPtr)smalloc(sizeof(int));
	if (!*r) return MallocFailError;
	**(int**)r = 1;
	return DAFSOK;
}

iDAFSError 
i_EndObjEdit(iEntityPtr ent,iEditRecordPtr r,int *deleteme)
{
	*deleteme = (*(int*)r > 1) && !ucslen(i_GetUniText(ent));
	smfree(r);
	return DAFSOK;
}

iDAFSError 
i_ObjEdit(iEntityPtr ent,iEditRecordPtr r,iUCode key,int *deleteme)
{
	iDAFSError e;
	iUCode buf[2];
	
	if (i_GetChildEntity(ent) && (e = i_Collapse(ent,0)))
		return e;
	*deleteme = 0;
	switch(key){
	case 20:	/*^u*/
		buf[0] = 0;
		if (e = i_SetUniText(ent,buf))
			return e;
		break;
		
	case '\b':
		if (*(int*)r == 1) {
			*deleteme = 1;
			return DAFSOK;
		}
		i_DelUni(ent,-1);
		break;
		
	default:
		if (*(int*)r > 1) i_AddUni(ent,-1,key);
		else {
			buf[0] = key;
			buf[1] = 0;
			if (e = i_SetUniText(ent,buf))
				return e;
		}
		break;
	}
	*(int*)r = 2;
	return DAFSOK;
	
}



/*
// bogus -- does not read in UTF
*/
iDAFSError 
i_Text2Dafs(char *text, iEntityPtr *dp)
{
	iEntityPtr doc,pl,pr;
	int beg,end;
	iDAFSError e;
	iEntityPtr dummy_nl;
	
	*dp = 0;
	doc = i_NewEntity(iDocumentT,0);
	if (!doc) return MallocFailError;
	dummy_nl = pl = pr = i_NewEntity(iNewLineT,doc);
	if (!pr) {
		i_DisposeEntity(doc);
		return MallocFailError;
	}
	i_SetInt(pl,1,0);
	beg = end = 0;
	while (*text) {
		/* bogus check here if it *text is a character or
		   a UTF sequence
		   */
		iUCode ubuf[4];
		int incr;
		incr = THOmbtouc(ubuf, text, 4);
		if (e = i_InsertChar(&pl,&beg,&pr,&end,ubuf[0])) {
			i_DisposeEntity(doc);
			return e;
		}
		text += incr;
	}
	*dp = doc;
	
	i_DisposeEntity(dummy_nl);

	return DAFSOK;
}

iDAFSError 
i_UCode2Dafs(iUCode *text,iEntityPtr *dp, int swap)
{
	iEntityPtr doc,pl,pr;
	int beg,end;
	iDAFSError e;
	iEntityPtr dummy_nl;
	
	*dp = 0;
	doc=i_NewEntity(iDocumentT,0);
	if (!doc) return MallocFailError;
	dummy_nl = pl = pr = i_NewEntity(iNewLineT,doc);
	if (!pr) {
		i_DisposeEntity(doc);
		return MallocFailError;
	}
	i_SetInt(pl,1,0);
	beg = end = 0;
	while (*text) {
		if (swap)
			*text = i_SwapS(*text);
		if (e = i_InsertChar(&pl,&beg,&pr,&end,*text)) {
			i_DisposeEntity(doc);
			return e;
		}
		text++;
	}
	*dp = doc;

	i_DisposeEntity(dummy_nl);

	return DAFSOK;
}
