/*
 * Copyright (c) 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */


/*******************************************************************************
 *  Contains the following routines for external use:
 *
 *  i_MakeARegex, i_MakeURegex, i_CheckARegex, i_CheckURegex,
 *  i_WriteRegex, i_ReadRegex, i_DisposeRegex
 ******************************************************************************* 
 *  Based on code which originally carried this notice:
 *
 *	Copyright (c) 1986 by University of Toronto.
 *	Written by Henry Spencer.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. The author is not responsible for the consequences of use of
 *		this software, no matter how awful, even if they arise
 *		from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *		by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software.
 *
 * Beware that some of this code is subtly aware of the way operator
 * precedence is structured in regular expressions.  Serious changes in
 * regular-expression syntax might require a total rethink.
 ******************************************************************************
 * Modifications and Notes:
 *
 * *** NOTE: this code has been altered slightly for use in 
 *           dafslib.
 *
 * *** NOTE: this code has been altered to handle character strings
 *           in Unicode representation. (Scott Small)
 *
 * *** NOTE: this code has been rearranged and renamed for inclusion in
 *           dafslib.  (Scott Small)
 ******************************************************************************/

#include "rafport.h"
#include "str-lib.h"

#include "ucs.h"
#include "utf.h"
#include "smemory.h"
#include "dafstype.h"
#include "entity.h"

#include "dafs.h"
#include "uregexp.h"
#include "rafutil.h"

/*
 * The "internal use only" fields in uregexp.h are present to pass info from
 * compile to execute that permits the execute phase to run lots faster on
 * simple cases.  They are:
 *
 * regstart	char that must begin a match; '\0' if none obvious
 * reganch	is the match anchored (at beginning-of-line only)?
 * regmust	string (pointer into program) that match must include, or NULL
 * regmlen	length of regmust string
 *
 * Regstart and reganch permit very fast decisions on suitable starting points
 * for a match, cutting down the work a lot.  Regmust permits fast rejection
 * of lines that cannot possibly match.  The regmust tests are costly enough
 * that MakeUniRegExpr() supplies a regmust only if the r.e. contains something
 * potentially expensive (at present, the only such thing detected is * or +
 * at the start of the r.e., which can involve a lot of backup).  Regmlen is
 * supplied because the test in CheckUniRegExpr() needs it and MakeUniRegExpr() is
 * computing it anyway.
 */

/*
 * Structure for regexp "program".  This is essentially a linear encoding
 * of a nondeterministic finite-state machine (aka syntax charts or
 * "railroad normal form" in parsing technology).  Each node is an opcode
 * plus a "next" pointer, possibly plus an operand.  "Next" pointers of
 * all nodes except BRANCH implement concatenation; a "next" pointer with
 * a BRANCH on both ends of it is connecting two alternatives.  (Here we
 * have one of the subtle syntax dependencies:  an individual BRANCH (as
 * opposed to a collection of them) is never concatenated with anything
 * because of operator precedence.)  The operand of some types of node is
 * a literal string; for others, it is a node leading into a sub-FSM.  In
 * particular, the operand of a BRANCH node is the first node of the branch.
 * (NB this is *not* a tree structure:  the tail of the branch connects
 * to the thing following the set of BRANCHes.)  The opcodes are:
 */

/* definition	number	opnd?	meaning */
#define	END	0	/* no	End of program. */
#define	BOL	1	/* no	Match "" at beginning of line. */
#define	UEOL	2	/* no	Match "" at end of line. */
#define	ANY	3	/* no	Match any one character. */
#define	ANYOF	4	/* str	Match any character in this string. */
#define	ANYBUT	5	/* str	Match any character not in this string. */
#define	BRANCH	6	/* node	Match this alternative, or the next... */
#define	BACK	7	/* no	Match "", "next" ptr points backward. */
#define	EXACTLY	8	/* str	Match this string. */
#define	NOTHING	9	/* no	Match empty string. */
#define	STAR	10	/* node	Match this (simple) thing 0 or more times. */
#define	PLUS	11	/* node	Match this (simple) thing 1 or more times. */
#define	OPEN	20	/* no	Mark this point in input as start of #n. */
			/*	OPEN+1 is number 1, etc. */
#define	CLOSE	30	/* no	Analogous to OPEN. */

/*
 * Opcode notes:
 *
 * BRANCH	The set of branches constituting a single choice are hooked
 *		together with their "next" pointers, since precedence prevents
 *		anything being concatenated to any individual branch.  The
 *		"next" pointer of the last BRANCH in a choice points to the
 *		thing following the whole choice.  This is also where the
 *		final "next" pointer of each individual branch points; each
 *		branch starts with the operand node of a BRANCH node.
 *
 * BACK		Normal "next" pointers all implicitly point forward; BACK
 *		exists to make loop structures possible.
 *
 * STAR,PLUS	'?', and complex '*' and '+', are implemented as circular
 *		BRANCH structures using BACK.  Simple cases (one character
 *		per match) are implemented with STAR and PLUS for speed
 *		and to minimize recursive plunges.
 *
 * OPEN,CLOSE	...are numbered at compile time.
 */

/*
 * A node is one char of opcode followed by two chars of "next" pointer.
 * "Next" pointers are stored as two 8-bit pieces, high order first.  The
 * value is a positive offset from the opcode of the node containing it.
 * An operand, if any, simply follows the node.  (Note that much of the
 * code generation knows about this implicit relationship.)
 *
 * NOTE: This data structure was modified to handle Unicode. The result
 *       is somewhat awkward, and could stand to be re-designed from
 *       scratch.
 *
 * Using two bytes for the "next" pointer is vast overkill for most things,
 * but allows patterns to get big without disasters.
 */
#define	OP(p)	   ( *((iUCode *)(p)) )
#define	NEXT(p)	   ( *((iUCode *)(p) + 1))
#define	OPERAND(p) ((iUCode *)(((iUCode *)(p)) + 2))

/*
 * Utility definitions.
 */
#define	iUCodeAT(p)	(*((iUCode *)(p)))

#define	FAIL(m)	{  return(NULL); }
#define	ISMULT(c)	((c) == '*' || (c) == '+' || (c) == '?')
#define	AMETA	"^$.[()|?+*\\"

/*
 * Flags to be passed up and down.
 */
#define	HASWIDTH	01	/* Known never to match null string. */
#define	SIMPLE		02	/* Simple enough to be STAR/PLUS operand. */
#define	SPSTART		04	/* Starts with * or +. */
#define	WORST		0	/* Worst case. */

/*
 * Global work variables for MakeUniRegExpr().
 */
static iUCode *regparse;		/* Input-scan pointer. */
static int2   regnpar;		/* () count. */
static iUCode regdummy;
static iUCode *regcode;		/* Code-emit pointer; &regdummy = don't. */
static long  regsize;		/* Code size. */
static iUCode META[12];          /* iUCode version of AMETA */

/*
 * The first byte of the regexp internal "program" is actually this magic
 * number; the start node begins in the second byte.
 */
#define	MAGIC	0234


/*
 * Forward declarations for MakeUniRegExpr()'s friends.
 */
static iUCode *reg(int paren, int *flagp);
static iUCode *regbranch(int *flagp);
static iUCode *regpiece(int *flagp);
static iUCode *regatom(int *flagp);
static iUCode *regnode(iUCode op);
static iUCode *regnext(iUCode *p);
static void  regc(iUCode b);
static void reginsert(iUCode op, iUCode *opnd);
static void regtail(iUCode *scan, iUCode *val);
static void regoptail(iUCode *p, iUCode *val);
static int ustrcspn(iUCode *s1, iUCode *s2);


/*
 * Global work variables.
 */
static iUCode *reginput;		/* String-input pointer. */
static iUCode *regbol;		/* Beginning of input, for ^ check. */
static iUCode **regstartp;	/* Pointer to startp array. */
static iUCode **regendp;		/* Ditto for endp. */

/*
 * Forwards.
 */
static int regtry(uregexp *prog, iUCode *string);
static int regmatch(iUCode *prog);
static int regrepeat(iUCode *p);
static int4 MatchUniRegExpr(void *data, iUCode *string, int4 *start, int4 *end);
static iDAFSError MakeUniRegExpr(iUCode *exp, void **data, char **err, uint2 *size);


/* ------------------- Public DAFSLib Routines ------------------ */
#if 0
static void propstr(iUCode *prog)
{
	ucsprint(OPERAND(prog));
}

static int prregop(iUCode **prog)
{
	iUCode op=OP(*prog);
	int retval=1;
	
	switch(op) {
		case END:
			retval=0;
			break;

		case BOL:
			printf("BOL ");
			break;

		case UEOL:
			printf("EOL ");
			break;

		case ANY:
			printf("ANY ");
			break;

		case ANYOF:
			printf("ANYOF ");
			propstr(*prog);
			break;

		case ANYBUT:
			printf("ANYBUT ");
			propstr(*prog);
			break;
		
		case BRANCH:
			printf("BRANCH ");
			propnode(*prog);
			break;
		
		case BACK:
			printf("BACK ");
			/* next goes backward - special case? */
			break;
		
		case EXACTLY:
			printf("EXACTLY ");
			propstr(*prog);
			break;
		
		case NOTHING:
			printf("NOTHING ");
			break;
		
		case STAR:
			printf("STAR ");
			propnode(*prog);
			break;
		
		case PLUS:
			printf("PLUS ");
			propnode(*prog);
			break;
		
		case OPEN:
			printf("OPEN ");
			break;
		
		case CLOSE:
			printf("CLOSE ");
			break;
		
		default:
			printf("Unknown OP Code");
			break;
	}
	return retval;
	
}

static void prreg(iRegexPtr r)
{
	iUCode *prog=r->prog;
	
	while (prregop(&prog))
		;
}
#endif

Dll iDAFSError APIENTRY
i_MakeARegex(iRegexPtr *rep, const char *str)
{
   /* Cover routine for i_MakeURegex, used when the expression *str
      is an ASCII string */

   iUCode      *ustr;
   iDAFSError retval;

   ustr = (iUCode *) smalloc((strlen(str)+1)*sizeof(iUCode));
   if (ustr == NULL) {
      return MallocFailError;
   } else {
	   utf2ucs(ustr, (char *) str);
      retval = i_MakeURegex(rep, ustr);
      smfree(ustr);        /* free iUCode string */
      return retval;
   }
} 

Dll int4 APIENTRY
i_CheckARegex(const iRegexPtr re, const char *str, int4 *start, int4 *end)
{
   /* Cover for i_CheckURegex, used when the string to be checked is ASCII */

   iUCode *ustr;
   int4   retval;

   ustr = (iUCode *) smalloc((strlen(str)+1)*sizeof(iUCode));
   if (ustr == NULL) {
      return 0;  /* no good way to indicate malloc fail error */
   } else {
	   utf2ucs(ustr, (char *) str);
      retval = i_CheckURegex(re, ustr, start, end);
      smfree(ustr);        /* free iUCode string */
      return retval;
   }
}

Dll iDAFSError APIENTRY
i_MakeURegex(iRegexPtr *rep, const iUCode *str)
{
   char *err;

   /* Allocate space for the iRegex structure and call MakeUniRegExpr */

   *rep =(iRegexPtr)smalloc(sizeof(iRegex));
   if (*rep == NULL)
      return MallocFailError;
   else
      return MakeUniRegExpr((iUCode*)str, &((*rep)->data), &err, &((*rep)->size));
}

Dll iDAFSError APIENTRY
i_WriteRegexFile(const iRegexPtr re, RafFile file)
{

	/* write out the specific data */
	return i_WriteItem(file, Data, re->size, re->data);
}


Dll iDAFSError APIENTRY
i_WriteRegex(const iRegexPtr re, char *filename)
{
	/* Copy the compiled regular expression to a RafFile */

	iDAFSError  e;
	RafFile        file;

	/* create the file and associated header */
	file = riOpen(filename,"wb");
	if (!file) return WriteLockedFileError;

	WriteHeader(file,DAFSB,0);

	e=i_WriteRegexFile(re,file);

	riClose(file);
	return e;

}

Dll iDAFSError APIENTRY
i_ReadRegexFile(char *type,iRegexPtr *rep, RafFile file, int *swap)
{
	iToken     t;
	iDAFSError e;
	iRegexPtr re;

	/* allocate  and initialize the iRegex structure */
	*rep = NULL;
	re = (iRegexPtr)smalloc(sizeof(iRegex));
	if (re == NULL) return MallocFailError;

	e = i_ReadToken(file, &t, swap);
	if (!e) e = i_ReadItemAlloc(file, t, &(re->size), &(re->data), swap);

	if (!e && *swap) {
		uregexp *rgp;
		rgp = (uregexp *) re->data;
		rgp->regmlen = i_SwapS(rgp->regmlen);   
	}
	if (e) {
		if (re->data)
			smfree(re->data);
		smfree(re);
	} else
		*rep = re;
	return e;
}

Dll iDAFSError APIENTRY
i_ReadRegex(iRegexPtr *rep, char *filename)
{
	/* Read a compiled regular expression from a file.  This routine
	   allocates space for the iRegex structure and the regular
	   expression itself. */

	iDAFSError e;
	RafFile       file;
	iDAFSWriteMode mode;
	int swap=0;

	file = riOpen(filename,"rb");
	if (!file)
		return NonExistentFileError;
	e = ReadHeader(file,&mode);
	if (e || mode!=DAFSB) {
		if (mode==DAFSB)
			e=BadFileType;
		riClose(file);
		return e;
	}

	if (e=i_ReadRegexFile("reg exp",rep,file,&swap)) return e;
	return DAFSOK;
}

Dll int4 APIENTRY
i_CheckURegex(const iRegexPtr re, const iUCode *str, int4 *start, int4 *end)
{
   /* Check a Unicode string for a substring matching the regular expression.
      NOTE:  if a match is found, *end is set so that str[*end-1] is the last
             character of the matching substring. */

   return MatchUniRegExpr((re)->data,(iUCode*) str, start, end); 
}

Dll void APIENTRY
i_DisposeRegex(iRegexPtr re)
{
   /* free the compiled regular expression, then the header */

   smfree(re->data);
   smfree(re);
}


/* ---------------------------- Static Routines ------------------------------- */

static iDAFSError
MakeUniRegExpr(iUCode *exp,void **data,char **err,uint2 *size)
/*
 * We can't allocate space until we know how big the compiled form will be,
 * but we can't compile it (and thus know how big it is) until we've got a
 * place to put the code.  So we cheat:  we compile it twice, once with code
 * generation turned off and size counting turned on, and once "for real".
 * This also means that we don't allocate space until we are sure that the
 * thing really will compile successfully, and we never have to move the
 * code and thus invalidate pointers into it.  (Note that it has to be in
 * one piece because free() must be able to free it all.)
 *
 * Beware that the optimization-preparation code in here knows about some
 * of the structure of the compiled regexp.
 */

{
	uregexp *r;
	iUCode *scan;
	iUCode *longest;
	int len;
	int flags;

	char *ameta = AMETA;
	utf2ucs(META, ameta);

	if (data) *data = 0;
	if (size) *size = 0;
	if (exp == NULL) {
		if (err) *err = "NULL argument";
		return DAFSOK;
	}
	/* First pass: determine size, legality. */
	regparse = exp;
	regnpar = 1;
	regsize = 0L;
	regcode = &regdummy;
	regc(MAGIC);
	if (reg(0, &flags) == NULL){
		if(err) *err = "Failed to Parse Regular Expresion";
		return CantParseString;
	}
	/* Small enough for pointer-storage convention? */
	if (regsize+sizeof(uregexp) >= MAX_SMALLOC)	{
		if (err) *err = "regexp too big";
		return UnknownDAFSError;
	}

	/* Allocate space. */
	r = (uregexp *)smalloc(sizeof(uregexp) + (unsigned)regsize);
	if (r == NULL)
		return MallocFailError;
	memset(r,0,sizeof(uregexp) + (unsigned)regsize);
	if (size) *size = (uint2)(sizeof(uregexp) + regsize);

	/* Second pass: emit code. */
	regparse = exp;
	regnpar = 1;
	regcode = r->program;
	regc(MAGIC);
	if (reg(0, &flags) == NULL){
		if(err) *err="Failed to Parse Regular Exprssion";
		return CantParseString;
	}
	/* Dig out information for optimizations. */
	r->regstart = '\0';	/* Worst-case defaults. */
	r->reganch = 0;
	r->regmust = NULL;
	r->regmlen = 0;
	scan = (iUCode *) (r->program+1);	/* First BRANCH. */
	if (OP(regnext(scan)) == END) {		/* Only one top-level choice. */
		scan = OPERAND(scan);

		/* Starting-point info. */
		if (OP(scan) == EXACTLY)
			r->regstart = *OPERAND(scan);
		else if (OP(scan) == BOL)
			r->reganch++;

		/*
		 * If there's something expensive in the r.e., find the
		 * longest literal string that must appear and make it the
		 * regmust.  Resolve ties in favor of later strings, since
		 * the regstart check works with the beginning of the r.e.
		 * and avoiding duplication strengthens checking.  Not a
		 * strong reason, but sufficient in the absence of others.
		 */
		if (flags&SPSTART) {
			longest = NULL;
			len = 0;
			for (; scan != NULL; scan = regnext(scan))
				if (OP(scan) == EXACTLY && ((int) ucslen(OPERAND(scan))) >= len) {
					longest = OPERAND(scan);
					len = ucslen(OPERAND(scan));
				}
			r->regmust = longest;
			r->regmlen = len;
		}
	}

	if (data) *data = (void*)r;
	return DAFSOK;
}


static int4
MatchUniRegExpr(void *data, iUCode *string,int4 *start,int4 *end)
{
	uregexp *prog = (uregexp*)data;
	iUCode *s;

	if (start) *start = 0;
	if (end) *end = 0;
	/* Check validity of program. */
	if (iUCodeAT(prog->program) != MAGIC) {
/*		RegError("corrupted program"); */
		return 0;
	}

	/* If there is a "must appear" string, look for it. */
	if (prog->regmust != NULL) {
		s = string;
		while ((s = (iUCode*)ucschr(s, prog->regmust[0])) != NULL) {
			if (ucsncmp(s, (iUCode *)prog->regmust, prog->regmlen) == 0)
				break;	/* Found it. */
			s++;
		}
		if (s == NULL)	/* Not present. */
			return 0;
	}

	/* Mark beginning of line for ^ . */
	regbol = string;

	/* Simplest case:  anchored match need be tried only once. */
	if (prog->reganch) {
		if (!regtry(prog, string)) return 0;
		if (end) *end = prog->endp[0]-string;
		return 1;
	}

	/* Messy cases:  unanchored match. */
	s = string;
	if (prog->regstart != '\0')
		/* We know what char it must start with. */
		while ((s = (iUCode*)ucschr(s, prog->regstart)) != NULL) {
			if (regtry(prog, s)) {
				if (start) *start = s - string;
				if (end) *end = prog->endp[0] - string;
				return 1;
			}
			s++;
		}
	else
		/* We don't -- general case. */
		do {
			if (regtry(prog, s)) {
				if (start) *start = s - string;
				if (end) *end = prog->endp[0] - string;
				return 1;
			}
		} while (*s++ != '\0');

	/* Failure. */
	return 0;
}




/*
 - reg - regular expression, i.e. main body or parenthesized thing
 *
 * Caller must absorb opening parenthesis.
 *
 * Combining parenthesis handling with the base level of regular expression
 * is a trifle forced, but the need to tie the tails of the branches to what
 * follows makes it hard to avoid.
 */
static iUCode *reg(int paren, int *flagp)
{
	iUCode *ret;
	iUCode *br;
	iUCode *ender;
	iUCode parno = 0;
	int flags;

	*flagp = HASWIDTH;	/* Tentatively. */

	/* Make an OPEN node, if parenthesized. */
	if (paren) {
		if (regnpar >= NSUBEXP/2)
			FAIL("too many ()");
		parno = regnpar;
		regnpar++;
		ret = regnode(OPEN+parno);
	} else
		ret = NULL;

	/* Pick up the branches, linking them together. */
	br = regbranch(&flags);
	if (br == NULL)
		return(NULL);
	if (ret != NULL)
		regtail(ret, br);	/* OPEN -> first. */
	else
		ret = br;
	if (!(flags&HASWIDTH))
		*flagp &= ~HASWIDTH;
	*flagp |= flags&SPSTART;
	while (*regparse == '|') {
		regparse++;
		br = regbranch(&flags);
		if (br == NULL)
			return(NULL);
		regtail(ret, br);	/* BRANCH -> BRANCH. */
		if (!(flags&HASWIDTH))
			*flagp &= ~HASWIDTH;
		*flagp |= flags&SPSTART;
	}

	/* Make a closing node, and hook it on the end. */
	ender = regnode((paren) ? CLOSE+parno : END);	
	regtail(ret, ender);

	/* Hook the tails of the branches to the closing node. */
	for (br = ret; br != NULL; br = regnext(br))
		regoptail(br, ender);

	/* Check for proper termination. */
	if (paren && *regparse++ != ')') {
		FAIL("unmatched ()");
	} else if (!paren && *regparse != '\0') {
		if (*regparse == ')') {
			FAIL("unmatched ()");
		} else
			FAIL("junk on end");	/* "Can't happen". */
		/* NOTREACHED */
	}

	return(ret);
}

/*
 - regbranch - one alternative of an | operator
 *
 * Implements the concatenation operator.
 */
static iUCode *regbranch(int *flagp)
{
	iUCode *ret;
	iUCode *chain;
	iUCode *latest;
	int flags;

	*flagp = WORST;		/* Tentatively. */

	ret = regnode(BRANCH);
	chain = NULL;
	while (*regparse != '\0' && *regparse != '|' && *regparse != ')') {
		latest = regpiece(&flags);
		if (latest == NULL)
			return(NULL);
		*flagp |= flags&HASWIDTH;
		if (chain == NULL)	/* First piece. */
			*flagp |= flags&SPSTART;
		else
			regtail(chain, latest);
		chain = latest;
	}
	if (chain == NULL)	/* Loop ran zero times. */
		(void) regnode(NOTHING);

	return(ret);
}

/*
 - regpiece - something followed by possible [*+?]
 *
 * Note that the branching code sequences used for ? and the general cases
 * of * and + are somewhat optimized:  they use the same NOTHING node as
 * both the endmarker for their branch list and the body of the last branch.
 * It might seem that this node could be dispensed with entirely, but the
 * endmarker role is not redundant.
 */
static iUCode *regpiece(int *flagp)
{
	iUCode *ret;
	iUCode op;
	iUCode *next;
	int flags;

	ret = regatom(&flags);
	if (ret == NULL)
		return(NULL);

	op = *regparse;
	if (!ISMULT(op)) {
		*flagp = flags;
		return(ret);
	}

	if (!(flags&HASWIDTH) && op != '?')
		FAIL("*+ operand could be empty");
	*flagp = (op != '+') ? (WORST|SPSTART) : (WORST|HASWIDTH);

	if (op == '*' && (flags&SIMPLE))
		reginsert(STAR, ret);
	else if (op == '*') {
		/* Emit x* as (x&|), where & means "self". */
		reginsert(BRANCH, ret);			/* Either x */
		regoptail(ret, regnode(BACK));		/* and loop */
		regoptail(ret, ret);			/* back */
		regtail(ret, regnode(BRANCH));		/* or */
		regtail(ret, regnode(NOTHING));		/* null. */
	} else if (op == '+' && (flags&SIMPLE))
		reginsert(PLUS, ret);
	else if (op == '+') {
		/* Emit x+ as x(&|), where & means "self". */
		next = regnode(BRANCH);			/* Either */
		regtail(ret, next);
		regtail(regnode(BACK), ret);		/* loop back */
		regtail(next, regnode(BRANCH));		/* or */
		regtail(ret, regnode(NOTHING));		/* null. */
	} else if (op == '?') {
		/* Emit x? as (x|) */
		reginsert(BRANCH, ret);			/* Either x */
		regtail(ret, regnode(BRANCH));		/* or */
		next = regnode(NOTHING);		/* null. */
		regtail(ret, next);
		regoptail(ret, next);
	}
	regparse++;
	if (ISMULT(*regparse))
		FAIL("nested *?+");

	return(ret);
}

/*
 - regatom - the lowest level
 *
 * Optimization:  gobbles an entire sequence of ordinary characters so that
 * it can turn them into a single node, which is smaller to store and
 * faster to run.  Backslashed characters are exceptions, each becoming a
 * separate node; the code is simpler that way and it's not worth fixing.
 */
static iUCode *regatom(int *flagp)
{
	iUCode *ret;
	int flags;

	*flagp = WORST;		/* Tentatively. */

	switch (*regparse++) {
	case '^':
		ret = regnode(BOL);
		break;
	case '$':
		ret = regnode(UEOL);
		break;
	case '.':
		ret = regnode(ANY);
		*flagp |= HASWIDTH|SIMPLE;
		break;
	case '[': {
			iUCode clss;
			iUCode classend;

			if (*regparse == '^') {	/* Complement of range. */
				ret = regnode(ANYBUT);
				regparse++;
			} else
				ret = regnode(ANYOF);
			if (*regparse == ']' || *regparse == '-')
				regc(*regparse++);
			while (*regparse != '\0' && *regparse != ']') {
				if (*regparse == '-') {
					regparse++;
					if (*regparse == ']' || *regparse == '\0')
						regc('-');
					else {
						clss = iUCodeAT(regparse-2)+1;
						classend = iUCodeAT(regparse);
						if (clss > classend+1)
							FAIL("invalid [] range");
						for (; clss <= classend; clss++)
							regc(clss);
						regparse++;
					}
				} else
					regc(*regparse++);
			}
			regc('\0');
			if (*regparse != ']')
				FAIL("unmatched []");
			regparse++;
			*flagp |= HASWIDTH|SIMPLE;
		}
		break;
	case '(':
		ret = reg(1, &flags);
		if (ret == NULL)
			return(NULL);
		*flagp |= flags&(HASWIDTH|SPSTART);
		break;
	case '\0':
	case '|':
	case ')':
		FAIL("internal urp");	/* Supposed to be caught earlier. */
		/* NOTREACHED */
		break;
	case '?':
	case '+':
	case '*':
		FAIL("?+* follows nothing");
		/* NOTREACHED */
		break;
	case '\\':
		if (*regparse == '\0')
			FAIL("trailing \\");
		ret = regnode(EXACTLY);
		regc(*regparse++);
		regc('\0');
		*flagp |= HASWIDTH|SIMPLE;
		break;
	default: {
			int len;
			iUCode ender;

			regparse--;
			len = ustrcspn(regparse, META);
			if (len <= 0)
				FAIL("internal disaster");
			ender = *(regparse+len);
			if (len > 1 && ISMULT(ender))
				len--;		/* Back off clear of ?+* operand. */
			*flagp |= HASWIDTH;
			if (len == 1)
				*flagp |= SIMPLE;
			ret = regnode(EXACTLY);
			while (len > 0) {
				regc(*regparse++);
				len--;
			}
			regc('\0');
		}
		break;
	}

	return(ret);
}

/*
 - regnode - emit a node
 */
static iUCode *regnode(iUCode op)
{
	iUCode *ret;

	ret = regcode;
	if (ret == &regdummy) {
		regsize += 4;
		return(ret);
	}

	*ret = op;
	*(ret+1)='\0';	/* null NEXT pointer */
	regcode = ret+2;   /* operand goes here */

	return(ret);
}

/*
 - regc - emit (if appropriate) a two bytes of code
 */
static void regc(iUCode b)
{
	if (regcode != &regdummy)
		*regcode++ = b;
	else
		regsize += 2;  /* 2 bytes per uCode operand */
}

/*
 - reginsert - insert an operator in front of already-emitted operand
 *
 * Means relocating the operand.
 */
static void reginsert(iUCode op, iUCode *opnd)
{
	iUCode *src;
	iUCode *dst;
	iUCode *place;

	if (regcode == &regdummy) {
		regsize += 4;
		return;
	}

	src = regcode;
	regcode += 2;                 /* advance by 2 iUCode = 4 bytes */
	dst = regcode;
	while (src > opnd)    /* copy 4 bytes */
		*--dst = *--src;

	place = opnd;		      /* Op node, where operand used to be. */
	*place++ = op;
	src = place;
	*src++ = '\0';
}

/*
 - regtail - set the next-pointer at the end of a node chain
 */
static void regtail(iUCode *scan, iUCode *val)
{
	iUCode *temp;
	unsigned short offset;

	if (scan == &regdummy)
		return;

	/* Find last node. */
	for (;;) {
		temp = regnext(scan);
		if (temp == NULL)
			break;
		scan = temp;
	}

	if (OP(scan) == BACK)
		offset = scan - val;
	else
		offset = val - scan;

	*(scan+1) = offset;
	
#if 0
	*(scan+2) = (offset>>8)&0xFF;
	*(scan+3) = offset&0xFF;
#endif

}

/*
 - regoptail - regtail on operand of first argument; nop if operandless
 */
static void regoptail(iUCode *p, iUCode *val)
{
	/* "Operandless" and "op != BRANCH" are synonymous in practice. */
	if (p == NULL || p == &regdummy || OP(p) != BRANCH)
		return;
	regtail(OPERAND(p), val);
}

/*
 - regtry - try match at specific point
 */
/* 0 failure, 1 success */
static int regtry(uregexp *prog, iUCode *string)
{
	int i;
	iUCode **sp;
	iUCode **ep;

	reginput = string;
	regstartp = (iUCode **)prog->startp;
	regendp = (iUCode **)prog->endp;

	sp = prog->startp;
	ep = prog->endp;
	for (i = NSUBEXP; i > 0; i--) {
		*sp++ = NULL;
		*ep++ = NULL;
	}
	if (regmatch(prog->program + 1)) {
		prog->startp[0] = string;
		prog->endp[0] = reginput;
		return(1);
	} else
		return(0);
}

/*
 - regmatch - main matching routine
 *
 * Conceptually the strategy is simple:  check to see whether the current
 * node matches, call self recursively to see whether the rest matches,
 * and then act accordingly.  In practice we make some effort to avoid
 * recursion, in particular by going through "ordinary" nodes (that don't
 * need to know whether the rest of the match failed) by a loop instead of
 * by recursion.
 */
/* 0 failure, 1 success */
static int regmatch(iUCode *prog)
{
	iUCode *scan;	/* Current node. */
	iUCode *next;		/* Next node. */   
        iUCode opcode; 

	scan = prog;
	while (scan != NULL) {
		next = regnext(scan);

		opcode = OP(scan);
		switch (opcode) {
		case BOL:
			if (reginput != regbol)
				return(0);
			break;
		case UEOL:
			if (*reginput != '\0')
				return(0);
			break;
		case ANY:
			if (*reginput == '\0')
				return(0);
			reginput++;
			break;
		case EXACTLY: {
				int2 len;
				iUCode *opnd;

				opnd = OPERAND(scan);
				/* Inline the first character, for speed. */
				if (*opnd != *reginput)
					return(0);
				len = ucslen(opnd);
				if (len > 1 && ucsncmp(opnd, reginput, len) != 0)
					return(0);
				reginput += len;
			}
			break;
		case ANYOF:
 			if (*reginput == '\0' || ucschr(OPERAND(scan), *reginput) == NULL)
				return(0);
			reginput++;
			break;
		case ANYBUT:
 			if (*reginput == '\0' || ucschr(OPERAND(scan), *reginput) != NULL)
				return(0);
			reginput++;
			break;
		case NOTHING:
			break;
		case BACK:
			break;
		case OPEN+1:
		case OPEN+2:
		case OPEN+3:
		case OPEN+4:
		case OPEN+5:
		case OPEN+6:
		case OPEN+7:
		case OPEN+8:
		case OPEN+9: {
				int no;
				iUCode *save;

				no = OP(scan) - OPEN;
				save = reginput;

				if (regmatch(next)) {
					/*
					 * Don't set startp if some later
					 * invocation of the same parentheses
					 * already has.
					 */
					if (regstartp[no] == NULL)
						regstartp[no] = save;
					return(1);
				} else
					return(0);
			}
			/* NOTREACHED */
			break;
		case CLOSE+1:
		case CLOSE+2:
		case CLOSE+3:
		case CLOSE+4:
		case CLOSE+5:
		case CLOSE+6:
		case CLOSE+7:
		case CLOSE+8:
		case CLOSE+9: {
				int no;
				iUCode *save;

				no = OP(scan) - CLOSE;
				save = reginput;

				if (regmatch(next)) {
					/*
					 * Don't set endp if some later
					 * invocation of the same parentheses
					 * already has.
					 */
					if (regendp[no] == NULL)
						regendp[no] = save;
					return(1);
				} else
					return(0);
			}
			/* NOTREACHED */
			break;
		case BRANCH: {
				iUCode *save;

				if (OP(next) != BRANCH)		/* No choice. */
					next = OPERAND(scan);	/* Avoid recursion. */
				else {
					do {
						save = reginput;
						if (regmatch(OPERAND(scan)))
							return(1);
						reginput = save;
						scan = regnext(scan);
					} while (scan != NULL && OP(scan) == BRANCH);
					return(0);
					/* NOTREACHED */
				}
			}
			/* NOTREACHED */
			break;
		case STAR:
		case PLUS: {
				iUCode nextch;
				int no;
				iUCode *save;
				int min;

				/*
				 * Lookahead to avoid useless match attempts
				 * when we know what character comes next.
				 */
				nextch = '\0';
				if (OP(next) == EXACTLY)
					nextch = *OPERAND(next);
				min = (OP(scan) == STAR) ? 0 : 1;
				save = reginput;
				no = regrepeat(OPERAND(scan));
				while (no >= min) {
					/* If it could work, try it. */
					if (nextch == '\0' || *reginput == nextch)
						if (regmatch(next))
							return(1);
					/* Couldn't or didn't -- back up. */
					no--;
					reginput = save + no;
				}
				return(0);
			}
			/* NOTREACHED */
			break;
		case END:
			return(1);	/* Success! */
			/* NOTREACHED */
			break;
		default:
/*			RegError("memory corruption"); */
			return(0);
			/* NOTREACHED */
			break;
		}

		scan = next;
	}

	/*
	 * We get here only if there's trouble -- normally "case END" is
	 * the terminating point.
	 */
/*	RegError("corrupted pointers"); */
	return(0);
}

/*
 - regrepeat - repeatedly match something simple, report how many
 */
static int regrepeat(iUCode *p)
{
	int count = 0;
	iUCode *scan;
	iUCode *opnd;

	scan = reginput;
	opnd = OPERAND(p);
	switch (OP(p)) {
	case ANY:
		count = ucslen(scan);
		scan += count;
		break;
	case EXACTLY:
		while (*opnd == *scan) {
			count++;
			scan++;
		}
		break;
	case ANYOF:
		while (*scan != '\0' && ucschr(opnd, *scan) != NULL) {
			count++;
			scan++;
		}
		break;
	case ANYBUT:
		while (*scan != '\0' && ucschr(opnd, *scan) == NULL) {
			count++;
			scan++;
		}
		break;
	default:		/* Oh dear.  Called inappropriately. */
/*		RegError("internal foulup"); */
		count = 0;	/* Best compromise. */
		break;
	}
	reginput = scan;

	return(count);
}

/*
 - regnext - dig the "next" pointer out of a node
 */
static iUCode *regnext(iUCode *p)
{
	int offset;

	if (p == &regdummy)
		return(NULL);

	offset = NEXT(p);
	if (offset == 0)
		return(NULL);

	if (OP(p) == BACK)
		return (p-offset);
	else
		return (p+offset);
}


/*
 * strcspn - find length of initial segment of s1 consisting entirely
 * of characters not from s2
 * NOTE: renamed and modified to handle strings of iUCode characters.  (SGS)
 */
static int ustrcspn(iUCode *s1, iUCode *s2)
{
	iUCode *scan1;
	iUCode *scan2;
	int count;

	count = 0;
	for (scan1 = s1; *scan1 != '\0'; scan1++) {
		for (scan2 = s2; *scan2 != '\0';)	/* ++ moved down. */
			if (*scan1 == *scan2++)
				return(count);
		count++;
	}
	return(count);
}
