
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
#include "rafport.h"
#include "str-lib.h"

#include <X11/keysym.h>

#include "KeyMap.h"

#define ESC 27

class Arabic : public KeyMap
{
public:
	Arabic(void);
	uString context(const uString &line);
	int spacing(iUCode character);
	void getInfo(int *direction, iUCode *letter);
};

Arabic::Arabic(void) 
	: KeyMap("Arabic")
{
	return;
}

#define NONE           0
#define RIGHT          1
#define LEFT           2
#define DUAL           3
#define TRANSPARENT    4
#define RIGHT_CAUSING  8
#define LEFT_CAUSING  16

// final, initial, medial
static iUCode forms[224][3] = {
	{ 0, 0, 0 },                   // 0x0620
	{ 0, 0, 0 },
	{ 0x82, 0, 0 },
	{ 0x84, 0, 0 },
	{ 0x86, 0, 0 },
	{ 0x88, 0, 0 },	
	{ 0x8C, 0x8A, 0x8B },
	{ 0x8E, 0, 0 },

	{ 0x92, 0x90, 0x91 },          // 0x0628 BAA
	{ 0x94, 0, 0 },
	{ 0x98, 0x96, 0x97 },
	{ 0x9C, 0x9A, 0x9B },	
	{ 0xA0, 0x9E, 0x9F },
	{ 0xA4, 0xA2, 0xA3 },
	{ 0xA8, 0xA6, 0xA7 },
	{ 0xAA, 0, 0 },

	{ 0xAC, 0, 0 },                // 0x0630 THAL
	{ 0xAE, 0, 0 },	
	{ 0xB0, 0, 0 },
	{ 0xB4, 0xB2, 0xB3 },
	{ 0xB8, 0xB6, 0xB7 },
	{ 0xBC, 0xBA, 0xBB },
	{ 0xC0, 0xBE, 0xBF },
	{ 0xC4, 0xC2, 0xC3 },	

	{ 0xC8, 0xC6, 0xC7 },          // 0x0638 DHAH
	{ 0xCC, 0xCA, 0xCB },
	{ 0xD0, 0xCE, 0xCF },
	{ 0, 0, 0 },
	{ 0, 0, 0 },
	{ 0, 0, 0 },
	{ 0, 0, 0 },
	{ 0, 0, 0 },

	{ 0, 0, 0 },                   // 0x0640 TATWEEL
	{ 0xD4, 0xD2, 0xD3 },
	{ 0xD8, 0xD6, 0xD7 },
	{ 0xDC, 0xDA, 0xDB },
	{ 0xE0, 0xDE, 0xDF },
	{ 0xE4, 0xE2, 0xE3 },
	{ 0xE8, 0xE6, 0xE7 },
	{ 0xEC, 0xEA, 0xEB },

	{ 0xEE, 0, 0 },	               // 0x0648 WAW
	{ 0xF0, 0, 0 },	
	{ 0xF4, 0xF2, 0xF3 }
};

static unsigned char linking[224];

static void 
formTables(void){
	memset(linking, 0, 224);

	for(int i = 0x0; i < 0x2A; i++){
		if(forms[i][1])
			linking[i] = RIGHT_CAUSING | LEFT_CAUSING | RIGHT | LEFT;
		else if(forms[i][0])
			linking[i] = LEFT_CAUSING | RIGHT;
	}
			
	linking[0x20] = RIGHT_CAUSING | LEFT_CAUSING;
	
	for(i = 0x2B; i <= 0x35; i++)
		linking[i] = TRANSPARENT;
}
	
uString
Arabic::context(const uString &line)
{
	unsigned i, length;
	uString glyphs;

	int cur_linking, next_linking, 
		prev_linking = NONE;

	// determine contextual forms
	i = 0;
	length = line.length();
	while(i < length){
		iUCode glyph = line[i];
		uString marks;
		
		if((glyph & 0xFF00) == 0x0600){
			// determine next base character
			unsigned j = 1;
			while(linking[line[i+j]-0x0620] & TRANSPARENT){
				/* handle the cases where a font exists to represent
				   a base character and a transparent mark */
				iUCode mark = line[i+j];

				if(glyph == 0x0627 && mark == 0x0653)       // maddah on alef
					glyph = 0x0622;
				else if(glyph == 0x0627 && mark == 0x0654)  // hamzah on alef
					glyph = 0x0623;
				else if(glyph == 0x0627 && mark == 0x0655)  // hamzah below alef
					glyph = 0x0625;
				else if(glyph == 0x0648 && mark == 0x0654)  // hamzah on waw
					glyph = 0x0624;
				else if(glyph == 0x064A && mark == 0x0654)  // hamzah on ya
					glyph = 0x0626;	
				else 
					marks += mark;

				j++;
			}
			
			// determine form
			if(line[i+j])
				next_linking = linking[line[i+j]-0x0620];
			else
				next_linking = NONE;
			cur_linking  = linking[glyph-0x0620];
			
			int form = NONE;
			if(cur_linking & RIGHT && prev_linking & RIGHT_CAUSING)
				form |= RIGHT;
			if(cur_linking & LEFT && next_linking & LEFT_CAUSING)
				form |= LEFT;
			prev_linking = cur_linking;			
			
			// handle ligatures
			if(glyph == 0x0644 && line[i+j] == 0x0627){
				if(form == DUAL)
					glyph = (iUCode) 0xFEFC;
				else
					glyph = (iUCode) 0xFEFB;
				
				j++;
				while(linking[line[i+j]-0x0620] & TRANSPARENT) j++;
				prev_linking = form = NONE;
			}			

			if(form) {
				uString temp((forms[glyph-0x0620][form-1] + 0xFE00));
				glyphs += temp;
			} else
				glyphs += glyph;
			glyphs +=  marks;

			i += j;
		}
		else{
			glyphs += glyph;
			prev_linking = NONE;
			i++;
		}		
	}


	return glyphs;
}	

int 
Arabic::spacing(iUCode character)
{
	if(character >= 0x064B && character <= 0x0655)
		return 0;

	if(character == 0xFEFB || character == 0xFEFC)
		return 2;

	return KeyMap::spacing(character);
}

void
Arabic::getInfo(int *direction, iUCode *letter)
{
	*direction = REVERSE;
	*letter = 0x0640;
}

KeyMap *
initArabic()
{
    Arabic *arabic;
    KeyMap *keymap;
	
	formTables();
    arabic = new Arabic;
    keymap = arabic;

        keymap->setTable(XK_quoteleft, 0x0639);
	keymap->setTable(XK_1, 0x0661);
	keymap->setTable(XK_2, 0x0662);
	keymap->setTable(XK_3, 0x0663);
	keymap->setTable(XK_4, 0x0664);
	keymap->setTable(XK_5, 0x0665);
	keymap->setTable(XK_6, 0x0666);
	keymap->setTable(XK_7, 0x0667);
	keymap->setTable(XK_8, 0x0668);
	keymap->setTable(XK_9, 0x0669);
	keymap->setTable(XK_0, 0x0660);

	keymap->setTable(XK_q, 0x0642);
	keymap->setTable(XK_w, 0x0648);
	keymap->setTable(XK_r, 0x0631);
	keymap->setTable(XK_T, 0x0637);
	keymap->setTable(XK_t, 0x062a);
	keymap->setTable(XK_y, 0x064a);
	keymap->setTable(XK_o, 0x0629);

	keymap->setTable(XK_A, 0x0649);
	keymap->setTable(XK_a, 0x0627);
	keymap->setTable(XK_S, 0x0635);
	keymap->setTable(XK_s, 0x0633);
	keymap->setTable(XK_D, 0x0636);
	keymap->setTable(XK_d, 0x062F);
	keymap->setTable(XK_f, 0x0641);
	keymap->setTable(XK_G, 0x063a);
	keymap->setTable(XK_H, 0x062d);
	keymap->setTable(XK_h, 0x0647);
	keymap->setTable(XK_j, 0x062c);
	keymap->setTable(XK_K, 0x062e);
	keymap->setTable(XK_k, 0x0643);
	keymap->setTable(XK_l, 0x0644);
	keymap->setTable(XK_quoteright, 0x0621);

	keymap->setTable(XK_Z, 0x0638);
	keymap->setTable(XK_z, 0x0632);
	keymap->setTable(XK_x, 0x0630);
	keymap->setTable(XK_c, 0x062b);
	keymap->setTable(XK_b, 0x0628);
	keymap->setTable(XK_n, 0x0646);
	keymap->setTable(XK_m, 0x0645);
	keymap->setTable(XK_comma, 0x060c);
	keymap->setTable(XK_period, '.');
	keymap->setTable(XK_question, 0x061f);
	keymap->setTable(XK_slash, 0x0634);
	
    keymap->setTable(XK_Delete, uString("Del"), '\b', uString());
	
    keymap->setTable(XK_Tab, uString("Tab"), '\t', uString('\t'));
    keymap->setTable(XK_Return, uString("Ret"), '\n', uString('\n'));
	
    keymap->setTable(XK_Control_L, uString("Control"), '\0', uString());
	
    keymap->setTable(XK_Shift_L, uString("Shift"), '\0', uString());
    keymap->setTable(XK_Shift_R, uString("Shift"), '\0', uString());
    keymap->setTable(XK_Linefeed, uString("LF"), '\0', uString());
	
    keymap->setTable(XK_Caps_Lock, uString(), '\0', uString());
    keymap->setTable(XK_Alt_L, uString("Alt"), '\0', uString());
    keymap->setTable(XK_Meta_L, uString("Meta"), '\0', uString());
    keymap->setTable(XK_space, uString(), ' ', uString(' '));
    keymap->setTable(XK_Meta_R, uString("Meta"), '\0', uString());
    keymap->setTable(XK_Alt_R, uString(), '\0', uString());
    keymap->setTable(XK_Super_R, uString(), '\0', uString());
	
    keymap->setTable(XK_Control_R, uString(), '\0', uString("xo"));

    keymap->setControlTable(XK_comma, 0x0655);
    keymap->setControlTable(XK_quoteright, 0x0654);
    keymap->setControlTable(XK_asciitilde, 0x0653);
	
    return keymap;
}


