
/*
 * Copyright (c) 1995 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "rafport.h"
#include "rafutil.h"

#include "smemory.h"

#include "Xm/Xm.h"
#include "uString.h"
#include "misc/Unicode.h"
#include "LineList.h"
#include "Line.h"
#include "utf.h"
#include "dafs.h"

LineList::LineList(void)
{
  size = 1;
  alloc = LINES_ALLOC;
  line = (Line **) smalloc(alloc * sizeof(Line *));
  line[0] = new Line(this);
}

LineList::~LineList(void)
{
  deleteAll();
  delete line[0];
  smfree(line);
}   

void
LineList::deleteAll(void)
{
  for(int i = 0; i < size; i++)
    delete line[i];
  line[0] = new Line(this);
  size = 1;
}

void
LineList::parseContext(void)
{
  for(int i = 0; i < size; i++)
    line[i]->parseContext();
}

Line *
LineList::nextLine(Line *prev)
{
  int i = 0;
  while(i < size-1)
    if(line[i++] == prev) 
      return line[i];

  return NULL;
}

Line *
LineList::prevLine(Line *next)
{
  int i = 1;
  while(i < size)
    if(line[i++] == next)
      return line[i-2];

  return NULL;
}

void
LineList::deleteLine(int index)
{
	int i;

  if(index >= 0 && index < size) {    
    for(i = line[index]->textLength(); i >= 0; i--)
      line[index]->removeChar(i);

    delete line[index];

    size--;
    for(i = index; i < size; i++)
      line[i] = line[i+1];
    
  }
}

void
LineList::detachLine(Line *that)
{
  int index = 0;
  while(index < size && line[index] != that)
    index++;

  delete line[index];
  
  size--;
  for(int i = index; i < size; i++)
    line[i] = line[i+1];
}

int
LineList::newLine(int index)
{
  if(index < 0) 
    index = 0;
  if(index > size)
    index = size;

  Line *new_line = new Line(this);
  if(!new_line)
    return FALSE;
  // bogus -- this is the place to use bcopy or memmove or some such
  if(index != size) {
    for(int i = size; i > index; i--)
      line[i] = line[i-1];
  }
  line[index] = new_line;
  size++;

  if(size == alloc) {
    alloc += LINES_ALLOC;
    line = (Line **) smrealloc(line, alloc * sizeof(Line *));
    if(!line) {
      delete line[index];
      size--;
      for(int i = index; i < size; i++)
	line[i] = line[i+1];
      return FALSE;
    }          
  }

  return TRUE;
}

// updates the lines new line info for the last line
// and then creates another line after it
int
LineList::newLine(iEntityPtr ent, int offset)
{
  line[size-1]->insertReference(line[size-1]->num_refs, ent, offset, 1);
  
  return newLine();
}

// splits line pos at index
void
LineList::splitLine(int pos, int index)
{
	int i;
  newLine(pos+1);

  Line *prev_line = line[pos], 
       *next_line = line[pos+1];

  if(prev_line->num_refs) {
    prev_line->insert((iUCode) '\n', index);
    index++;
    
    // copy references
    int reference = prev_line->reference(index);
    prev_line->splitReference(reference, index);
    reference = prev_line->reference(index);

    int next_text_length = 0;
    for(i = reference; i < prev_line->num_refs; i++){
      next_line->insertReference(i-reference, prev_line->refs[i].en,
				 prev_line->refs[i].en_index,
				 prev_line->referenceLength(i));
      next_line->refs[i-reference].text_index = next_text_length;
      next_text_length += prev_line->referenceLength(i);
    }

	// idea of this line: you have just split the line at
	// reference, and all the entities to the right 
	// should no longer have references on this line, so remove them

	// problem: if you have an entity with embedded newlines
	// this can remove the reference to the newline that
	// is supposed to stay -- I believe I have fixed this
	// by fixing the calculation of where to do the deleting
	for(i = prev_line->num_refs-1; i >= reference; i--)
	  prev_line->deleteReference(i);

    // copy new line info and create new newline
    index--;
    prev_line->text.remove(index);
  }
    
  // update text, context and indices
  next_line->text = prev_line->text.string()+index;
  prev_line->text.remove(index, 0xFFFF); // remove all to end
  prev_line->parseContext();
  next_line->parseContext();
}

void
LineList::splitUni(const iUCode *uni, iEntityPtr en)
{
  int length = ucslen(uni);
  iUCode *temp = (iUCode *) smalloc((length+1) * sizeof(iUCode));
  ucscpy(temp, uni);
  
  int offset = 0;
  iUCode *end, *start = temp;
  while(end = (iUCode *) ucschr(start, '\n')) {
    *end = 0;
    line[size-1]->addUnicode(start, en, offset);
    newLine(en, (int) (end - temp));
    start = end + 1;
    offset = (int) (start - temp);
  }
  if(start < temp+length || length == 0)
    line[size-1]->addUnicode(start, en, offset);

  smfree(temp);
}

void
LineList::parseEntityStub(iEntityPtr en)
{
  Line *line_on = line[size-1];
  int i;

  if(i_GetEntityType(en) == iSpaceT) {                // spaces
    int num_spaces = i_GetShort(en, NULL);
    iUCode *uni = (iUCode *) smalloc(sizeof(iUCode) * (num_spaces+1));
    for(i = 0; i < num_spaces; i++)
      uni[i] = ' ';
    uni[i] = 0;
    line_on->addUnicode(uni, en, 0);
  }
  else if(i_GetEntityType(en) == iNewLineT) {         // new lines
    int i, num_lines = i_GetShort(en, NULL);
    for(i = 0; i < num_lines; i++)
      newLine(en, i);
  }
  else switch(i_GetEDataType(en)) {
  case iDString: {                                    // ascii strings
    const char *ascii = i_GetText(en, NULL);
    iUCode *uni = (iUCode *) smalloc(sizeof(iUCode) * ucsMaxSize(ascii));
    utf2ucs(uni, ascii);
    splitUni(uni, en);
	smfree(uni);
    break;
  }

  case iDUString: {                                   // unicode strings
    const iUCode *uni = i_GetUText(en, NULL);
    splitUni(uni, en);
    break;
  }

  case iDPosSet:                                      // possibility sets
  case iD1PosSet: {
    static iUCode uni[2] = { 0, 0 };
    *uni = i_GetGlyph(en, NULL);
    if(*uni == '\n') 
      newLine(en, 0);
    else
      line_on->addUnicode(uni, en, 0);
    break;
  }
    
  default: {                                          // children 
    iEntityPtr child;
    for(child = i_GetChildEntity(en); child; child = i_GetNextEntity(child))
      parseEntityStub(child);
  }
  }
}

void
LineList::parseEntity(iEntityPtr root)
{
  parseEntityStub(root);

  // delete last line if it has no data
  // ie. parseEntityStub always creates a next line
  if(size == 1 && !line[0]->num_refs) { // no data in root
    iEntityPtr new_en = i_NewEntity(iNewLineT, root);
    i_SetShort(new_en, 1, 0);

    parseEntityStub(root);
  }

  if(line[size-1]->num_refs == 0)
    deleteLine(size-1);
}

static void
prlines(LineList *lines) 
{
  for(int i = 0; i < lines->size; i++) {
    Line *line = lines->line[i];
    
    char *ascii = (char *) smalloc(4*(line->text.length()+1));
    ucs2utf(ascii, line->text.string());
    printf("%s\n", ascii);
    for(int j = 0; j < line->num_refs; j++)
      printf("  %p (at %i) from %i\n", line->refs[j].en, 
	     line->refs[j].en_index, line->refs[j].text_index);
  }
}

static void 
pparse(iEntityPtr root)
{
  LineList lines;
  lines.parseEntity(root);
  prlines(&lines);
}




