
/*
 * Copyright (c) 1994 RAF Technology
 * developed for ARPA under contract MDA 904-93-C-7219
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * RAF Technology or ARPA  may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of RAF Technology and ARPA.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL RAF TECHNOLOGY OR ARPA  BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
/* Written by Dan Heller.  Copyright 1991, O'Reilly && Associates.
 * This program is freely distributable without licensing fees and
 * is provided without guarantee or warrantee expressed or implied.
 * This program is -not- in the public domain.
 */

/* build_option.c -- The final version of BuildMenu() is used to
 * build popup, option, pulldown -and- pullright menus.  Menus are
 * defined by declaring an array of MenuItem structures as usual.
 */
#include <Xm/MainW.h>
#include <Xm/PanedW.h>
#include <Xm/RowColumn.h>
#include <Xm/DrawingA.h>
#include <Xm/CascadeBG.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/ScrolledW.h>

#include "misc.h"

/* Build popup, option and pulldown menus, depending on the menu_type.
 * It may be XmMENU_PULLDOWN, XmMENU_OPTION or  XmMENU_POPUP.  Pulldowns
 * return the CascadeButton that pops up the menu.  Popups return the menu.
 * Option menus are created, but the RowColumn that acts as the option
 * "area" is returned unmanaged. (The user must manage it.)
 * Pulldown menus are built from cascade buttons, so this function
 * also builds pullright menus.  The function also adds the right
 * callback for PushButton or ToggleButton menu items.
 */
Widget
BuildMenu(Widget parent,
		  int menu_type,
		  char *menu_title, 
		  char menu_mnemonic,
		  MenuItem *items)
{
    Widget menu, cascade, widget;
    int i;
    XmString str;
	
    if (menu_type == XmMENU_PULLDOWN || menu_type == XmMENU_OPTION)
        menu = XmCreatePulldownMenu(parent, "_pulldown", NULL, 0);
    else if (menu_type == XmMENU_POPUP)
        menu = XmCreatePopupMenu(parent, "_popup", NULL, 0);
    else {
        XtWarning("Invalid menu type passed to BuildMenu()");
        return NULL;
    }
	
    /* Pulldown menus require a cascade button to be made */
    if (menu_type == XmMENU_PULLDOWN) {
        str = XmStringCreateLocalized(menu_title);
        cascade = XtVaCreateManagedWidget(menu_title,
										  xmCascadeButtonGadgetClass, parent,
										  XmNsubMenuId,   menu,
										  XmNlabelString, str,
/*
// we want to get those meta events for the keyboard
*/
#if 0
										  XmNmnemonic,    menu_mnemonic,
#endif
										  NULL);
        XmStringFree(str);
    } else if (menu_type == XmMENU_OPTION) {
        /* Option menus are a special case, but not hard to handle */
        Arg args[2];
        str = XmStringCreateLocalized(menu_title);
        XtSetArg(args[0], XmNsubMenuId, menu);
        XtSetArg(args[1], XmNlabelString, str);
        /* This really isn't a cascade, but this is the widget handle
         * we're going to return at the end of the function.
         */
        cascade = XmCreateOptionMenu(parent, menu_title, args, 2);
        XmStringFree(str);
    }
	
    /* Now add the menu items */
    for (i = 0; items[i].label != NULL; i++) {
        /* If subitems exist, create the pull-right menu by calling this
         * function recursively.  Since the function returns a cascade
         * button, the widget returned is used..
         */
        if (items[i].subitems)
            if (menu_type == XmMENU_OPTION) {
                XtWarning("You can't have submenus from option menu items.");
                continue;
            } else
                widget = BuildMenu(menu, XmMENU_PULLDOWN,
								   items[i].label, items[i].mnemonic, items[i].subitems);
        else
			items[i].result = widget = XtVaCreateManagedWidget(items[i].label,
															   *items[i].klass, menu,
															   NULL);
		
        /* Whether the item is a real item or a cascade button with a
         * menu, it can still have a mnemonic.
         */

		
        /* any item can have an accelerator, except cascade menus. But,
         * we don't worry about that; we know better in our declarations.
         */
        if (items[i].accelerator) {
			str = XmStringCreateLocalized(items[i].accel_text);
			XtVaSetValues(widget,
						  XmNaccelerator, items[i].accelerator,
						  XmNacceleratorText, str,
						  NULL);
			XmStringFree(str);
        }
		
        if (items[i].callback)
			XtAddCallback(widget,
						  (items[i].klass == &xmToggleButtonWidgetClass ||
						   items[i].klass == &xmToggleButtonGadgetClass)?
						  XmNvalueChangedCallback : /* ToggleButton class */
						  XmNactivateCallback,      /* PushButton class */
						  items[i].callback, items[i].callback_data);
    }
	
    /* for popup menus, just return the menu; pulldown menus, return
     * the cascade button; option menus, return the thing returned
     * from XmCreateOptionMenu().  This isn't a menu, or a cascade button!
     */
    return menu_type == XmMENU_POPUP? menu : cascade;
}
Widget
BuildXmStringMenu(Widget parent,
		  int menu_type,
		  XmString menu_title, 
		  char menu_mnemonic,
		  XmFontList font_list,
		  XmStringMenuItem *items)
{
    Widget menu, cascade, widget;
    int i;
    XmString str;
	
    if (menu_type == XmMENU_PULLDOWN || menu_type == XmMENU_OPTION)
        menu = XmCreatePulldownMenu(parent, "_pulldown", NULL, 0);
    else if (menu_type == XmMENU_POPUP)
        menu = XmCreatePopupMenu(parent, "_popup", NULL, 0);
    else {
        XtWarning("Invalid menu type passed to BuildMenu()");
        return NULL;
    }
	
    /* Pulldown menus require a cascade button to be made */
    if (menu_type == XmMENU_PULLDOWN) {
        str = menu_title;
        cascade = XtVaCreateManagedWidget(menu_title,
										  xmCascadeButtonGadgetClass, parent,
										  XmNsubMenuId,   menu,
										  XmNlabelString, str,
										  NULL);
        /*XmStringFree(str);*/
    } else if (menu_type == XmMENU_OPTION) {
        /* Option menus are a special case, but not hard to handle */
        Arg args[2];
        str = XmStringCreateLocalized(menu_title);
        XtSetArg(args[0], XmNsubMenuId, menu);
        XtSetArg(args[1], XmNlabelString, str);
        /* This really isn't a cascade, but this is the widget handle
         * we're going to return at the end of the function.
         */
        cascade = XmCreateOptionMenu(parent, menu_title, args, 2);
        XmStringFree(str);
    }
    /* Now add the menu items */
    for (i = 0; items[i].label != NULL; i++) {
        /* If subitems exist, create the pull-right menu by calling this
         * function recursively.  Since the function returns a cascade
         * button, the widget returned is used..
         */
        if (items[i].subitems)
            if (menu_type == XmMENU_OPTION) {
                XtWarning("You can't have submenus from option menu items.");
                continue;
            } else
                widget = BuildXmStringMenu(menu, XmMENU_PULLDOWN,
								   items[i].label, items[i].mnemonic,
								   font_list,
								   items[i].subitems);
        else {
			items[i].result = widget = XtVaCreateManagedWidget(items[i].label,
															   *items[i].klass, menu,
															   XmNlabelString, items[i].label,
															   XmNfontList, font_list,
															   NULL);
		}
												   
		
        /* Whether the item is a real item or a cascade button with a
         * menu, it can still have a mnemonic.
         */
		
        /* any item can have an accelerator, except cascade menus. But,
         * we don't worry about that; we know better in our declarations.
         */
        if (items[i].accelerator) {
			str = XmStringCreateLocalized(items[i].accel_text);
			XtVaSetValues(widget,
						  XmNaccelerator, items[i].accelerator,
						  XmNacceleratorText, str,
						  NULL);
			XmStringFree(str);
        }
		
        if (items[i].callback)
			XtAddCallback(widget,
						  (items[i].klass == &xmToggleButtonWidgetClass ||
						   items[i].klass == &xmToggleButtonGadgetClass)?
						  XmNvalueChangedCallback : /* ToggleButton class */
						  XmNactivateCallback,      /* PushButton class */
						  items[i].callback, items[i].callback_data);
    }
	
    /* for popup menus, just return the menu; pulldown menus, return
     * the cascade button; option menus, return the thing returned
     * from XmCreateOptionMenu().  This isn't a menu, or a cascade button!
     */
    return menu_type == XmMENU_POPUP? menu : cascade;
}

