#include "VDD.h"
#include "RTEView.h"
#include "Rectangle.h"
#include "RTEMisc.h"
#include "RTEList.h"
#include "RTECursor.h"
#include "RTE.h"
#include "RootElement.h"
#include "smemory.h"
#include "compare.h"

#define DEBUG_RECTANGLE FALSE


// ---------------------------- Rectangle methods -----------------------------
Rectangle::Rectangle()
{
	x = y = w = h = 0;
	color = DEFAULT_RECTANGLE_COLOR;
	line_style = DEFAULT_RECTANGLE_LINE_STYLE;
}

Rectangle::Rectangle(Coord px, Coord py)
{
	x = px;  y = py;
	h = w = 0;
	color = DEFAULT_RECTANGLE_COLOR;
	line_style = DEFAULT_RECTANGLE_LINE_STYLE;
}

Rectangle::Rectangle(Coord px, Coord py, Coord pw, Coord ph)
{
	x = px;  y = py;
	w = pw;  h = ph;
	color = DEFAULT_RECTANGLE_COLOR;
	line_style = DEFAULT_RECTANGLE_LINE_STYLE;
}

int Rectangle::moveBy(Coord dx, Coord dy, int squeeze, int allow_overlap)
{
	return Rectangle::moveTo(this->getX() + dx, this->getY() + dy, squeeze, allow_overlap);
}

int Rectangle::moveTo(Coord x, Coord y, int squeeze, int allow_overlap)
{
	// squeeze and allow_overlap are ignored rectangles which are not children

	this->setX(x);
	this->setY(y);
	return TRUE;
}

int Rectangle::changeSizeTo(Coord w, Coord h, int redraw)
{
	// redraw is ignored for plain Rectangles

	this->setW(w);
	this->setH(h);
	return TRUE;
}

Rectangle::Rectangle(SimpleInterval *ix, SimpleInterval *iy)
{
	x = ix->getX();  y = iy->getX();
	w = ix->getDX(); h = iy->getDX();
	color = DEFAULT_RECTANGLE_COLOR;
	line_style = DEFAULT_RECTANGLE_LINE_STYLE;
}

void Rectangle::setColor(ColorType c)
{
	this->color = c;
}

void Rectangle::setLineStyle(LineStyle ls)
{
	this->line_style = ls;
}

Rectangle * Rectangle::intersection(Rectangle *r)
{
	// Assume r's coords are relative to my parent.
	// Construct rectangle with coords relative to my parent
	// representing the intersection of me and r.

	Rectangle *result = 0;

	// upper left and lower right corner of intersection
	Coord ulx = MAX(this->x, r->x),
      	      uly = MAX(this->y, r->y),
      	      lrx = MIN(this->x + this->w - 1, r->x + r->w - 1),
      	      lry = MIN(this->y + this->h - 1, r->y + r->h - 1);
	if(ulx <= lrx && uly <= lry)
		result = new Rectangle(ulx, uly, lrx-ulx+1, lry-uly+1);

	return result;
}

static void constructRectangleUnion(Rectangle *r1, Rectangle *r2, Rectangle *r)
{
	// set r to be the union of r1 and r2
	// assume r1 and r2 have coords relative to same parent

	// upper left and lower right corner of union
	const Coord ulx = MIN(r1->getX(), r2->getX()),
      	            uly = MIN(r1->getY(), r2->getY()),
      	            lrx = MAX(r1->getX() + r1->getW() - 1, r2->getX() + r2->getW() - 1),
      	            lry = MAX(r1->getY() + r1->getH() - 1, r2->getY() + r2->getH() - 1);

	r->setX(ulx);
	r->setY(uly);
	r->setW(lrx-ulx+1);
	r->setH(lry-uly+1);
}


Rectangle * Rectangle::areaUnion(Rectangle *r)
{
	// Assume r's coords are relative to my parent.
	// Construct minimal rectangle containing me and r.
	// Coords of result rectangle are relative to my parent.

	Rectangle *result = new Rectangle();
	if(result)
		constructRectangleUnion(this, r, result);
	
	return result;
}

Rectangle * Rectangle::areaUnion(ElementList *rl)
{
	Rectangle *result = NULL;
	if(rl->numElements() > 0) {
		Rectangle *r0 = rl->getItemNumber(0);
		result = new Rectangle(r0->getX(), r0->getY(), r0->getW(), r0->getH());
		if(result) {
			for(int i=1; i<rl->numElements(); i++) {
				if(rl->getItemNumber(i)->positionIsValid())
					constructRectangleUnion(rl->getItemNumber(i), result, result);
			}
		}
	}
	return result;
}

int Rectangle::removeIntersection(Rectangle *r, RectangleSet **result)
{
	// Assume r's coords are relative to my parent.
	// result = me - intersection(me,r) 
	// May return 0, 1, 3, or 5 Rectangles.
	// Coords of result rectangles are relative to my parent.

	int ret_val = TRUE + 1; // bogus initial value

	if(!(*result = new RectangleSet())) {
		ret_val = FALSE; // failure
	} else if(!this->overlaps(r)) {
		// result is a copy of me
		Rectangle *new_r = new Rectangle(x, y, w, h);
		if(new_r) {
			if((*result)->addElement(new_r))
				ret_val = TRUE; // success
			else {
				delete new_r;  new_r = 0;
				ret_val = FALSE;  // failure
			}
		}
	} else if(this->contains(r)) {
		// result is 1, 2, 3, or 4  Rectangles

		// if seperated by comma's - treated as compound expresion
		SimpleInterval *ix1 = new SimpleInterval(this->x, r->x - this->x);
		SimpleInterval *ix2 = new SimpleInterval(r->x, r->w );
		SimpleInterval *ix3 = new SimpleInterval(r->x + r->w - 1, this->x + this->w - (r->x + r->w));
		SimpleInterval *iy1 = new SimpleInterval(this->y, r->y - this->y);
		SimpleInterval *iy2 = new SimpleInterval(this->y, this->h);
		SimpleInterval *iy3 = new SimpleInterval(r->y + r->h - 1, this->y + this->h - (r->y + r->h));

		Rectangle *r1 = new Rectangle(ix1, iy2),
		          *r2 = new Rectangle(ix2, iy3),
		          *r3 = new Rectangle(ix3, iy2),
		          *r4 = new Rectangle(ix2, iy1);

		delete ix1; delete ix2; delete ix3; delete iy1; delete iy2; delete iy3;
		ix1 = ix2 = ix3 = iy1 = iy2 = iy3 = 0;

		ret_val = (r1 && r2 && r3 && r4);

		if(ret_val && r1->area() > 0) 
			ret_val = (*result)->addElement(r1);
		else { 
			delete r1; r1 = 0; 
		}

		if(ret_val && r2->area() > 0) 
			ret_val = (*result)->addElement(r2);
		else {
			delete r2; r2 = 0;
		}

		if(ret_val && r3->area() > 0) 
			ret_val = (*result)->addElement(r3);
		else {
			delete r3; r3 = 0;
		}

		if(ret_val && r4->area() > 0) 
			ret_val = (*result)->addElement(r4);
		else {
			delete r4; r4 = 0;
		}

	} else if(r->contains(this)) {
		// result is empty set
		ret_val = TRUE;
	} else {
		// partial overlap - trim Rectangle and call this routine
		// recursively (I will now contain r)
		Rectangle *r_trim = this->intersection(r);
		ret_val = this->removeIntersection(r_trim, result);
		delete r_trim; r_trim = 0;
	}

	assert(ret_val == TRUE || ret_val == FALSE);
	return ret_val;
}

int Rectangle::overlaps(Rectangle *r)
{
	// Assume r's coords are relative to my parent.
	// Determine whether I have any area in common with r.

	// upper left and lower right corner of intersection
	Coord ulx = MAX(this->x, r->x),
      	      uly = MAX(this->y, r->y),
      	      lrx = MIN(this->x + this->w - 1, r->x + r->w - 1),
      	      lry = MIN(this->y + this->h - 1, r->y + r->h - 1);
	return (ulx <= lrx && uly <= lry);
}

int Rectangle::isEqualTo(Rectangle *r)
{
	return ( this->getX() == r->getX() && this->getY() == r->getY() &&
	         this->getW() == r->getW() && this->getH() == r->getH() );
}

int Rectangle::contains(Coord px, Coord py)
{
	// Coords px and py are relative to my parent.
	// See if I contain the point (px,py)

	return ( this->x <= px  &&  px <= this->x + this->w - 1 &&
	         this->y <= py  &&  py <= this->y + this->h - 1);
}

int Rectangle::contains(Rectangle *r)
{
	// Assume r's coords are relative to my parent.
	// See if I completely contain r.

	return (this->x <= r->x && 
	        this->y <= r->y &&
	        this->x + this->w >= r->x + r->w &&
	        this->y + this->h >= r->y + r->h);
}

// -------------------------- ChildRectangle methods ----------------------------

ChildRectangle::ChildRectangle()
{
	parent = NULL;
	selected = FALSE;
	this->setPositionNotValid();
}

ChildRectangle::ChildRectangle(Coord x, Coord y):
	Rectangle(x, y)
{
	parent = NULL;
	selected = FALSE;
	this->setPositionNotValid();
}

ChildRectangle::ChildRectangle(Coord x, Coord y, Coord w, Coord h):
	Rectangle(x, y, w, h)
{
	parent = NULL;
	selected = FALSE;
	this->setPositionNotValid();
}
	
ChildRectangle::ChildRectangle(ParentRectangle *par,Coord x, Coord y, Coord w, Coord h):
	Rectangle(x, y, w, h)
{
	parent = par;
	selected = FALSE;
	this->setPositionNotValid();
}
	
	
void ChildRectangle::getRootOffset(Coord *pdx, Coord *pdy)
{
	*pdx += this->x;
	*pdy += this->y;
	this->parent->getRootOffset(pdx, pdy);
}


int ChildRectangle::positionIsValid()
{
	return (this->position_valid && this->parent->positionIsValid());
}

void ChildRectangle::getAbsoluteCoords(Coord *px, Coord *py)
{
	*px = *py = 0;
	this->getRootOffset(px, py);
}

void ChildRectangle::getRelativeCoords(Coord ax, Coord ay, Coord *rx, Coord *ry)
{
	*rx = *ry = 0;
	this->getRootOffset(rx, ry);
	*rx = ax - *rx;
	*ry = ay - *ry;
}

void ChildRectangle::safeDelete(int squeeze)
{
	// Send me to the deleted pool associated with the
	// root element above me.

	this->getRoot()->sendToDeleted(this);
	this->parent->deleteChild(this, squeeze);
}

void ChildRectangle::undelete()
{
	// Remove me from the deleted pool associated with the
	// root element above me.

	this->getRoot()->getFromDeleted(this);
	this->parent->registerChild(this);
}

int ChildRectangle::moveBy(Coord dx, Coord dy, int squeeze, int allow_overlap)
{
	return ChildRectangle::moveTo(this->getX() + dx, this->getY() + dy, squeeze, allow_overlap);
}

int ChildRectangle::moveTo(Coord new_x, Coord new_y, int squeeze, int allow_overlap)
{
	int result = TRUE;
	const Coord old_x = this->x, old_y = this->y;
	if(old_x != new_x || old_y != new_y) {
		const int redraw = TRUE, erase = TRUE;
		ParentRectangle *par = this->parent;
		if(par) {
			par->deleteChild(this, squeeze, erase);   // delete at old location
			this->parent = NULL;
		}
		this->setX(new_x);  this->setY(new_y);      // change location
		if(par) result = this->becomeChildOf(par, redraw, allow_overlap); // insert at new location
		if(!result) {
			this->setX(old_x);  this->setY(old_y);
			const int redraw = this->positionIsValid();
			if(par && !this->becomeChildOf(par, redraw))
				halt("unable to register child back at original location\n");
		}
	}
	return result;
}


int ChildRectangle::changeSizeTo(Coord new_w, Coord new_h, int redraw)
{
	int result = FALSE;
	if(new_w == this->w && new_h == this->h) {
		result = TRUE; // no change
	} else {
		// heinous - assumes all parents are packed
		PackedRectangle *par = (PackedRectangle *)this->parent;
		if(par && this->position_valid) {
			// see if our parent might let us grow at current location
			int i_was_changed = FALSE;
			const Coord old_x = this->x, old_y = this->y, old_w = this->w, old_h = this->h;
			Rectangle *tr = new Rectangle(this->x, this->y, this->w, this->h);
			par->changeChildSize(new_w, new_h, tr);
			if(par->canGrowToContain(tr)) {
				//const int squeeze = (new_w < this->getW());
				const int squeeze = FALSE;
				const int erase = redraw;
				ChildRectangle *pc = par->prevChild(this);
				par->deleteChild(this, squeeze, erase);
				this->parent = NULL;
				this->Rectangle::changeSizeTo(new_w, new_h, FALSE);
				// this->changeSizeTo(new_w, new_h, FALSE);
				i_was_changed = TRUE;
				#if 0
				if(pc)
					par->positionAfterChild(pc, this);
				else
					par->positionFirst(this);
				#endif
				result = this->becomeChildOf(par, redraw);
			}
			if(!result) {
				// see if our parent might take us at start of next line
				par->positionAtStartOfNextLine(tr);
				if(par->canGrowToContain(tr)) {
					if(!i_was_changed) {
						const int squeeze = FALSE, erase = TRUE;
						par->deleteChild(this, squeeze, erase);
						this->parent = NULL;
						par->changeChildSize(new_w, new_h, this);
						i_was_changed = TRUE;
					}
					// we moved, and hence must redraw
					const int redraw = TRUE;
					par->positionAtStartOfNextLine(this);
					result = this->becomeChildOf(par, redraw);
				}
			}
			delete tr;
			if(i_was_changed && !result) {
				// restore me to original location
				const int redraw = TRUE;
				this->setW(old_w);  this->setH(old_h);
				this->setX(old_x);  this->setY(old_y);
				if(!this->becomeChildOf(par, redraw))
					halt("unable to change child back to original size\n");
			}
		} else {
			// we are not registered to a parent, and are free to grow at will
			result = this->Rectangle::changeSizeTo(new_w, new_h);
		}
	}
	return result;
}

int ChildRectangle::changeSizeBy(Coord dw, Coord dh, int redraw)
{
	return this->changeSizeTo(this->getW() + dw, this->getH() + dh, redraw);
}

RootElement *ChildRectangle::getRoot()
{
	if(this->parent)
		return this->parent->getRoot();
	else
		return NULL;
}

int ChildRectangle::becomeChildOf(ParentRectangle *p, int redraw, int allow_overlap)
{
	int result;

	assert(p != NULL && !p->isParentOf(this));
	this->setPositionNotValid();
	this->parent = NULL;
	if(p->registerChild(this, allow_overlap)) {
		result = TRUE;   // success
		this->parent = p;
		this->setPositionValid();
	} else {
		result = FALSE;
	}

	assert(!result || p->isParentOf(this));
	if(result && redraw) this->requestRedraw();
	return result;
}

int
ChildRectangle::isParentOf(ChildRectangle *child)
{
  if(child) {
    ParentRectangle *parent = child->parent;
    
    while(parent && parent != this)
      parent = parent->parent;
    
    return (parent != NULL);
  }

  return 0;
}

void ChildRectangle::getLeafElementsAtLocation(Coord px, Coord py, ElementSet *s)
{
	assert(px >= 0 && py >= 0 && px < w && py < h);
	s->addElement(this);
}

void ChildRectangle::getLeafElementsInRectangle(Rectangle *r, ElementSet *s)
{
	s->addElement(this);
}

int ChildRectangle::drawBorder(void)
{
	RootElement *root = getRoot();
	if(root){
		VirtualDisplayDevice *vdd = root->getDisplayDevice();
		Coord x_pos, y_pos;
		this->getAbsoluteCoords(&x_pos, &y_pos);
		vdd->drawRect(x_pos, y_pos, w, h);
	}
	return 0;
}

int
ChildRectangle::erase(void)
{
 	RootElement *root = getRoot();
	if(root){
		VirtualDisplayDevice *vdd = root->getDisplayDevice();
		Coord x_pos, y_pos;
		this->getAbsoluteCoords(&x_pos, &y_pos);
		vdd->eraseRect(x_pos, y_pos, w, h);
	}
	return 0;
} 


// ------------------------------ ParentRectangle methods ----------------------------

ParentRectangle::ParentRectangle()
{
	h_interval_list = new IntervalList;
	v_interval_list = new IntervalList;
	element_list = new ElementList(compareCRCR);
}

ParentRectangle::ParentRectangle(Coord x, Coord y):
	ChildRectangle(x, y)
{
	h_interval_list = new IntervalList;
	v_interval_list = new IntervalList;
	element_list = new ElementList(compareCRCR);
}

ParentRectangle::ParentRectangle(Coord x, Coord y, Coord w, Coord h):
	ChildRectangle(x, y, w, h)
{
	h_interval_list = new IntervalList;
	v_interval_list = new IntervalList;
	element_list = new ElementList(compareCRCR);
}

ParentRectangle::~ParentRectangle()
{
	assert(h_interval_list != 0 && v_interval_list != 0 && element_list != 0);
	delete h_interval_list; h_interval_list = 0;
	delete v_interval_list; v_interval_list = 0;
	delete element_list;  element_list = 0;
}


int ParentRectangle::isParentOf(ChildRectangle *c) 
{ 
	return this->element_list->contains(c);
}

int ParentRectangle::contains(Rectangle *r)
{
	// Assume r is a child of me and see if r fits inside me 
	// at its current position relative to me.

	return (r->x >= 0 &&
	        r->y >= 0 &&
	        r->x + r->w <= this->w &&
	        r->y + r->h <= this->h);
}

void ParentRectangle::deleteAllChildren()
{
	for(int i=this->numChildren()-1; i >= 0; i--)
		this->deleteChild(this->getChildNumber(i), FALSE);
}

void ParentRectangle::deleteAllValidChildren()
{
	for(int i=this->element_list->numElements()-1; i >= 0; i--) {
		ChildRectangle *c = this->getChildNumber(i);
		if(c->positionIsValid())
			this->deleteChild(c, FALSE);
	}
}

Rectangle * ParentRectangle::areaUnion(Rectangle *r)
{
	// Construct rectangle whose coords are relative to my parent, representing
	// the union of me with r.  On input, r's coords are relative to me.
	// rtemp is the same as r, but with coords relative to my parent.

	Rectangle *rtemp = new Rectangle(r->x + this->x, r->y + this->y, r->w, r->h),
	          *result = this->Rectangle::areaUnion(rtemp);
	delete rtemp;
	return result;
}

Rectangle *ParentRectangle::intersection(Rectangle *r)
{
	// Construct rectangle whose coords are relative to me, representing
	// the intersection of me with r.
	
	Rectangle *rtemp = new Rectangle(r->x + this->x, r->y + this->y, r->w, r->h),
	          *result = this->Rectangle::intersection(rtemp);
	if(result) {
		result->setX(result->x - this->x);
		result->setY(result->y - this->y);
	}
	delete rtemp;
	return result;
}


int ParentRectangle::registerChild(ChildRectangle *c)
{
	const int allow_overlap = FALSE;
	return this->registerChild(c, allow_overlap);
}

int ParentRectangle::registerChild(ChildRectangle *c, int allow_overlap)
{
	int success=FALSE;

	if(allow_overlap) 
	  success = ((this->contains(c) || this->growToContain(c)) && this->registerAreaAllocation(c, c));
	else 
	  success = this->allocateArea(c, c);

	success = success && this->element_list->insertItem(c);

	assert(!success || this->isParentOf(c));
	return success;
}

void ParentRectangle::deleteChild(ChildRectangle *c, int squeeze, int erase)
{
	// squeeze is ignored for non-packed ParentRectangle

	assert(c->parent == this);
	if(erase) c->requestErase();
	this->freeArea(c);

	#ifndef NDEBUG
	ElementSet *my_children = this->getChildrenInRectangle(this);
	assert(!my_children->contains(c));
	delete my_children;
	#endif

	element_list->removeItemNumber(element_list->locationOfItem(c));
	c->parent = NULL;
	c->setPositionNotValid();
}

void ParentRectangle::positionOnTopOfChild(ChildRectangle *c, Rectangle *r)
{
	r->setX(c->getX());
	r->setY(c->getY());
}

ElementList * ParentRectangle::getAllChildren()
{
	ElementList *result = new ElementList(compareCRCR);
	if(!this->element_list->isEmpty()) 
		this->element_list->copyTo(result);
	return result;
}

ElementList * ParentRectangle::getAllValidChildren()
{
	ElementList *result = new ElementList(compareCRCR);
	if(!this->element_list->isEmpty()) 
		this->element_list->copyTo(result);
	for(int i=0; i<this->element_list->numElements(); i++) {
		ChildRectangle *c = this->element_list->getItemNumber(i);
		if(!c->positionIsValid())
			result->removeItem(c);
	}
	return result;
}

ElementSet *ParentRectangle::getChildrenAtLocation(Coord px, Coord py)
{
	ElementSet *v_hits, *h_hits = this->h_interval_list->findHit(px);
	if(h_hits)
	       v_hits = this->v_interval_list->findHit(py);

	if(h_hits && v_hits) h_hits->intersectWith(v_hits);
	if(v_hits) {
		delete v_hits; v_hits = 0;
	}
	return h_hits;
}

ElementSet *ParentRectangle::getChildrenInRectangle(Rectangle *r)
{
	Interval *h_interval = new Interval(r->getX(), r->getW());
	ElementSet *h_hits=0;
	if(h_interval) {
		h_hits = this->h_interval_list->findHit(h_interval);
		delete h_interval; h_interval = 0;
		if(h_hits && !h_hits->isEmpty()) {
	        	Interval *v_interval = new Interval(r->getY(), r->getH());
			if(v_interval) {
				ElementSet *v_hits = this->v_interval_list->findHit(v_interval);
				delete v_interval; v_interval = 0;
				if(v_hits) {
					h_hits->intersectWith(v_hits);
					delete v_hits; v_hits = 0;
				} else {
					delete h_hits; h_hits = 0;
				}
			}
		}
	}
	return h_hits;
}

ElementList *ParentRectangle::getListOfChildrenInRectangle(Rectangle *r)
{
	ElementSet *child_set = this->getChildrenInRectangle(r);
	ElementList *result_list = NULL;
	int success = TRUE;
	if(child_set && !child_set->isEmpty()) {
		if(result_list = new ElementList(compareCRCR))
			for(int i=0; success && i<child_set->numElements(); i++)
				success = result_list->insertItem(child_set->getElementNumber(i));
	}
	delete child_set;
	return result_list;
}

int ParentRectangle::changeSizeTo(Coord new_w, Coord new_h, int redraw)
{
	// See if I will still contain all my children at the new
	// size.
	int can_change = TRUE;
	#if 0
	if(new_w < this->getW() || new_h < this->getH()) {
		Rectangle *r = new Rectangle(0, 0, this->getW(), this->getH());
		if(this->parent)
			this->parent->changeChildSize(new_w, new_h, r);
		for(int i=0; can_change && i<this->element_list->getSize(); i++) {
			ChildRectangle *c = this->getChildNumber(i);
			if(c->positionIsValid())
				can_change = r->contains(c);
		}
		delete r;
	}
	#endif

	if(can_change)
		return this->ChildRectangle::changeSizeTo(new_w, new_h, redraw);
	else
		return FALSE;
}

void ParentRectangle::changeChildSize(Coord new_w, Coord new_h, Rectangle *child)
{
	// Make no attempt to keep child alligned with line spacing, since none exists.

	child->w = new_w;
	child->h = new_h;
}

int ParentRectangle::inOrder(ChildRectangle *a, ChildRectangle*b)
{
	// Compare children by their address (pointer value) since
	// no reading order exists for non-packed ParentRectangle.

	if((int)a < (int)b)
		return LESS_THAN;
	else if((int)a == (int)b)
		return EQUAL;
	else
		return GREATER_THAN;
}


void ParentRectangle::getLeafElementsAtLocation(Coord x, Coord y, ElementSet *leaves)
{
	ElementSet *children = this->getChildrenAtLocation(x, y);
	if(!children || children->isEmpty())  {
		// hit empty space in me
		leaves->addElement(this);
	} else { 
		for(int i=0; i < children->numElements(); i++) {
			ChildRectangle *e = children->getElementNumber(i);
			e->getLeafElementsAtLocation(x-e->getX(), y-e->getY(), leaves);
		}
	}
	delete children;
}

void ParentRectangle::getLeafElementsInRectangle(Rectangle *r, ElementSet  *leaves)
{
	ElementSet *children = this->getChildrenInRectangle(r);
	if(!children || children->isEmpty()) 
	        leaves->addElement(this);
	else {
	        for(int i=0; i < children->numElements(); i++) {
			ChildRectangle *e = children->getElementNumber(i);
			Rectangle *r_offset = new Rectangle(r->getX() - e->getX(),
			                      	r->getY() - e->getY(), r->getW(), r->getH());
			e->getLeafElementsInRectangle(r_offset, leaves);
			delete r_offset;
		}
	}
	delete children;
}

ElementSet * ParentRectangle::getLeafElementsAtLocation(Coord x, Coord y)
{
	// Search down the Rectangle hierarchy until we hit either a leaf element
	// or empty space within a ParentRectangle.  If we hit empty space, return
	// the lowest level parent containing the space.
	
	//  Allow for more than one (i.e. overlapping) leaf elements at one location.
	
	assert(x >= 0 && y >= 0);
	
	ElementSet *leaves = new ElementSet();
	this->getLeafElementsAtLocation(x, y, leaves);
	return leaves;
}

ElementSet * ParentRectangle::getLeafElementsInRectangle(Rectangle *r)
{
	ElementSet *leaves = new ElementSet();
	this->getLeafElementsInRectangle(r, leaves);
	return leaves;
}

int ParentRectangle::allocateArea(Coord x1, Coord y1, Coord x2, Coord y2, 
                                            ChildRectangle *cr)
{
	int success = FALSE;
	const Coord w = x2 - x1 + 1,  h = y2 - y1 + 1;
	ChildRectangle *requested_rect = new ChildRectangle(x1, y1, w, h);

	if(requested_rect) {
		success = this->allocateArea(requested_rect, cr);
		delete requested_rect; requested_rect = 0;
	}
	return success;
}

int ParentRectangle::allocateArea(ChildRectangle *r, ChildRectangle *cr, ElementList **conflict_list)
{
	ElementSet *conflict_set;
	const int success = this->allocateArea(r, cr, &conflict_set);
	if(conflict_set && !conflict_set->isEmpty()) {
		*conflict_list = new ElementList(compareCRCR);
		for(int i=0; i<conflict_set->numElements(); i++)
			(*conflict_list)->insertItem(conflict_set->getElementNumber(i));	
	} else  
		*conflict_list = NULL;
	if(conflict_set) delete conflict_set;
	return success;
}


int ParentRectangle::allocateArea(ChildRectangle *r, ChildRectangle *cr)
{
	ElementSet *conflicts;
	const int success = allocateArea(r, cr, &conflicts);
	if(conflicts) delete conflicts; 
	return success;
}

int ParentRectangle::allocateArea(ChildRectangle *r, ChildRectangle *cr, ElementSet **conflicts)
{
	// Assume r's coords are relative to me.

	int success;

	if(r->area() == 0)
		success = TRUE;  // null area requested
	else {
		success = FALSE;
		if(this->contains(r)) {
			*conflicts = this->getChildrenInRectangle(r);
			if(*conflicts) {
				(*conflicts)->removeElement(cr);
				success = (*conflicts)->isEmpty() &&
					this->registerAreaAllocation(r, cr);
				if(success) {
					delete *conflicts;
					*conflicts = 0;
				}
			}
		} else halt("attempt to allocate area outside parent\n");
	}
	return success;
}

int ParentRectangle::registerAreaAllocation(Rectangle *r, ChildRectangle *cr)
{
	// Assume r's coords are relative to me.

	assert(this->contains(r));
	Interval *h_int = new Interval(r->x, r->w, cr),
	         *v_int = new Interval(r->y, r->h, cr);

	int success = h_int && v_int &&	
	              this->h_interval_list->add(&h_int) &&
	              this->v_interval_list->add(&v_int);
				
	if(h_int) { delete h_int; h_int = 0; }
	if(v_int) { delete v_int; v_int = 0; }

	if(DEBUG_RECTANGLE) {
		debugf("In ParentRectangle::register_area_allocation:\n");
		h_interval_list->debugPrint("h_interval_list\n");
		v_interval_list->debugPrint("v_interval_list\n");
		debugf("\n");
	}
	return success;
}

int ParentRectangle::freeArea(ChildRectangle *cr)
{
	// Free the entire area allocated to cr.  
	// Assume that cr's x, y, w, h match its current area allocation.
	// Note:  this operation is much faster than freeing a partial area,
	// and should be used where possible.
	
	int success=FALSE;
	Interval *h_int = new Interval(cr->getX(), cr->getW(), cr),
	         *v_int = new Interval(cr->getY(), cr->getH(), cr);

	if(h_int && v_int) {
		this->h_interval_list->deleteElement(h_int, cr);
		this->v_interval_list->deleteElement(v_int, cr);
		success = TRUE;
	}
	if(h_int) { delete h_int; h_int = 0; }
	if(v_int) { delete v_int; v_int = 0; }

	return success;
}

int ParentRectangle::freeArea(Rectangle *r, ChildRectangle *cr)
{
	// First create intervals corresponding to the area to be freed (may
	// not be necessary, but we can't tell), then delete cr from those 
	// intervals.

	int success;

	if(r->area() == 0)
		success = TRUE;  // null area to free
	else {
		success = FALSE;
		Interval *h_int = new Interval(r->getX(), r->getW(), cr),
			 *v_int = new Interval(r->getY(), r->getH(), cr);
		if(h_int && v_int) {
			if(this->h_interval_list->add(&h_int) &&
			   this->v_interval_list->add(&v_int)) {
				// add function may have destroyed the intervals
				if(!h_int) h_int = new Interval(r->getX(), r->getW(), cr);
				if(!v_int) v_int = new Interval(r->getY(), r->getH(), cr);
				if(h_int && v_int) {
					this->h_interval_list->deleteElement(h_int, cr);
					this->v_interval_list->deleteElement(v_int, cr);
					success = TRUE;
				}
			}
		}
		if(h_int) { delete h_int; h_int = 0; }
		if(v_int) { delete v_int; v_int = 0; }
	}
	return success;
}


int ParentRectangle::freeArea(RectangleSet *s, ChildRectangle *cr)
{
	int result = TRUE;
	for(int i=0; i<s->numElements(); i++)
		result = result && this->freeArea(s->getElementNumber(i), cr);
	return result;
}

int ParentRectangle::draw()
{
	int result = TRUE;
	for(int i=0; i<this->element_list->getSize(); i++)
		this->element_list->getItemNumber(i)->requestRedraw();
	return result;
}

int ParentRectangle::erase()
{
	int result = TRUE;
	for(int i=0; i<this->element_list->getSize(); i++)
		this->element_list->getItemNumber(i)->requestErase();
	return result;
}

int ParentRectangle::drawRegion(Rectangle *rect)
{
	int result = TRUE;
	ElementSet *toDraw = this->getChildrenInRectangle(rect);
	if(toDraw) {
	  for(int i = 0; result && i < toDraw->numElements(); i++) 
	    toDraw->getElementNumber(i)->requestRedraw();
	}
	delete toDraw;
	return result;
}



#ifdef MIKE
extern "C" 
{
  static void
  prrectStub(ChildRectangle *rect, int count)
  {
    for(int i = 0; i < count; i++)
      fprintf(stderr, "  ");
    
    char type[256];
    switch(rect->getType()) {
    case UCRectangleT:
      strcpy(type, "UCRectangle");
      break;
    case UCStringT:
      strcpy(type, "UCString");
      break;
    case UCParaT:
      strcpy(type, "UCPara");
      break;
    case RootElementT:
      strcpy(type, "RootElement");
      break;
    case CursorT:
      strcpy(type, "Cursor");
      break;
    default:
      sprintf(type, "Type: %i", rect->getType());
      break;
    }

    fprintf(stderr, "%p - %s %4i, %4i, %4i, %4i\n", rect, type, rect->getX(), rect->getY(), rect->getW(), rect->getH());
    
    if(rect->numChildren()) {
      ChildRectangle *child;
      for(child = ((PackedRectangle *) rect)->firstChild(); 
	  child; 
	  child = ((PackedRectangle *) rect)->nextChild(child))
	prrectStub(child, count+1);
    }
  }
    
  void
  prrect(ChildRectangle *rect)
  {
    prrectStub(rect, 0);
  }
}
#endif
